# pylint: disable=unused-import

from mau.lexers.main_lexer import MainLexer
from mau.lexers.text_lexer import TextLexer
from mau.nodes.page import ContainerNode, DocumentNode
from mau.parsers.base_parser import ParserError
from mau.parsers.environment import Environment
from mau.parsers.main_parser import MainParser
from mau.text_buffer.context import Context
from mau.text_buffer.text_buffer import TextBuffer
from mau.visitors.base_visitor import BaseVisitor

# from mau.visitors.html_visitor import HTMLVisitor
# from mau.visitors.markua_visitor import MarkuaVisitor
# from mau.visitors.latex_visitor import LatexVisitor


# pylint: disable=import-outside-toplevel
def load_visitors():
    """
    This function loads all the visitors belonging to
    the group "mau.visitors". This code has been isolated
    in a function to allow visitor modules to import the
    Mau package without creating a cycle.
    """

    import sys

    if sys.version_info < (3, 10):
        from importlib_metadata import entry_points
    else:
        from importlib.metadata import entry_points

    discovered_plugins = entry_points(group="mau.visitors")

    # Load the available visitors
    visitors = [i.load() for i in discovered_plugins]
    visitors.append(BaseVisitor)

    return visitors


class ConfigurationError(ValueError):
    """Used to signal an error in the configuration"""


class Mau:
    def __init__(
        self,
        input_file_name,
        visitor_class,
        config=None,
        custom_templates=None,
        templates_directory=None,
        full_document=False,
    ):
        # This is needed to set up the initial context for the lexer
        self.input_file_name = input_file_name

        # This is the class that implements the selected visitor
        self.visitor_class = visitor_class

        # This is a configuration dictionary that
        # will be used here and also injected into the variables
        self.config = config or {}

        # A dictionary with the custom templates.
        self.custom_templates = custom_templates

        # A directory that contains the custom templates as files
        self.templates_directory = templates_directory

        # If this flag is True the output is wrapped in a DocumentNode,
        # otherwise Mau will use a ContainerNode
        self.full_document = full_document

        # This will contain all the variables declared
        # in the text and in the configuration
        self.environment = Environment()

    def run_lexer(self, text):
        context = Context(source=self.input_file_name)
        text_buffer = TextBuffer(text, context)

        lexer = MainLexer(text_buffer)
        lexer.process()

        return lexer

    def run_parser(self, tokens):
        # Create the Mau environment
        self.environment.update(self.config, namespace="mau")

        # Parse the source text using the given configuration
        parser = MainParser(tokens, environment=self.environment)
        parser.parse()

        # Create the footnotes
        parser.create_footnotes()

        # Process references
        parser.process_references()

        return parser

    def process(self, nodes, environment):
        wrapper_node_class = ContainerNode
        if self.full_document:
            wrapper_node_class = DocumentNode

        # Wrap the whole output
        output = wrapper_node_class(nodes)

        # Initialise the visitor
        # Use the parser variables so that the visitor
        # has both the configuration values and the
        # variables defined inside the text
        visitor = self.visitor_class(
            custom_templates=self.custom_templates,
            templates_directory=self.templates_directory,
            config=environment.asdict(),
        )

        # Replace variables with the ones processed by the parser
        # self.variables = parser.variables

        # Visit the document AST
        return visitor.visit(output)



from mau.nodes.nodes import Node
from mau.nodes.page import PageNode


class ReferenceNode(Node):
    """Reference to stored content."""

    node_type = "reference"

    def __init__(
        self,
        content_type,
        name,
        category=None,
        content=None,
        number=None,
        title=None,
        reference_anchor=None,
        content_anchor=None,
    ):
        super().__init__()
        self.content_type = content_type
        self.name = name
        self.category = category
        self.content = content or []
        self.number = number
        self.title = title
        self.reference_anchor = reference_anchor
        self.content_anchor = content_anchor

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "content_type": self.content_type,
            "name": self.name,
            "category": self.category,
            "content": self.content,
            "number": self.number,
            "title": self.title,
            "reference_anchor": self.reference_anchor,
            "content_anchor": self.content_anchor,
        }


class ReferencesEntryNode(ReferenceNode):
    """An entry of the list of references."""

    node_type = "references_entry"


class CommandReferencesNode(PageNode):
    """This instructs Mau to insert the content of references."""

    node_type = "command_references"

    def __init__(
        self,
        entries,
        content_type=None,
        category=None,
        name=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(args, kwargs, tags)
        self.content_type = content_type
        self.category = category
        self.name = name
        self.entries = entries

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "content_type": self.content_type,
            "category": self.category,
            "name": self.name,
            "entries": self.entries,
            "args": self.args,
            "kwargs": self.kwargs,
            "tags": self.tags,
        }



from mau.nodes.inline import TextNode, VerbatimNode
from mau.nodes.references import (
    CommandReferencesNode,
    ReferenceNode,
    ReferencesEntryNode,
)


def test_reference_node():
    node = ReferenceNode(
        "somecontent",
        "somevalue",
        content=[VerbatimNode("somevalue"), TextNode("othervalue")],
    )

    assert node.content_type == "somecontent"
    assert node.name == "somevalue"
    assert node.category is None
    assert node.content == [VerbatimNode("somevalue"), TextNode("othervalue")]
    assert node.number is None
    assert node.title is None
    assert node.reference_anchor is None
    assert node.content_anchor is None
    assert node.node_type == "reference"
    assert node == ReferenceNode(
        "somecontent",
        "somevalue",
        content=[VerbatimNode("somevalue"), TextNode("othervalue")],
    )


def test_reference_node_with_number_and_anchor():
    node = ReferenceNode(
        "somecontent",
        "somevalue",
        "somecategory",
        [VerbatimNode("somevalue"), TextNode("othervalue")],
        number="3",
        title="Exercise 1",
        reference_anchor="someanchor",
        content_anchor="someanchor-def",
    )

    assert node.content_type == "somecontent"
    assert node.name == "somevalue"
    assert node.category == "somecategory"
    assert node.content == [VerbatimNode("somevalue"), TextNode("othervalue")]
    assert node.number == "3"
    assert node.title == "Exercise 1"
    assert node.reference_anchor == "someanchor"
    assert node.content_anchor == "someanchor-def"
    assert node.node_type == "reference"
    assert node == ReferenceNode(
        "somecontent",
        "somevalue",
        "somecategory",
        [VerbatimNode("somevalue"), TextNode("othervalue")],
        number="3",
        title="Exercise 1",
        reference_anchor="someanchor",
        content_anchor="someanchor-def",
    )


def test_references_entry_node():
    node = ReferencesEntryNode(
        "somecontent",
        "somevalue",
        "somecategory",
        content=TextNode("Some text"),
        number="1",
        title="Exercise 1",
        reference_anchor="someanchor",
        content_anchor="someanchor-def",
    )

    assert node.content_type == "somecontent"
    assert node.name == "somevalue"
    assert node.category == "somecategory"
    assert node.content == TextNode("Some text")
    assert node.number == "1"
    assert node.title == "Exercise 1"
    assert node.reference_anchor == "someanchor"
    assert node.content_anchor == "someanchor-def"
    assert node.node_type == "references_entry"
    assert node == ReferencesEntryNode(
        "somecontent",
        "somevalue",
        "somecategory",
        content=TextNode("Some text"),
        number="1",
        title="Exercise 1",
        reference_anchor="someanchor",
        content_anchor="someanchor-def",
    )


def test_command_references_node():
    node = CommandReferencesNode(
        content_type="somecontent",
        name="somevalue",
        category="somecategory",
        entries=[TextNode("somevalue1"), TextNode("somevalue2")],
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )

    assert node.entries == [TextNode("somevalue1"), TextNode("somevalue2")]
    assert node.content_type == "somecontent"
    assert node.name == "somevalue"
    assert node.category == "somecategory"
    assert node.args == ["value1", "value2"]
    assert node.tags == ["tag1", "tag2"]
    assert node.kwargs == {"key1": "text1", "key2": "text2"}
    assert node.node_type == "command_references"
    assert node == CommandReferencesNode(
        content_type="somecontent",
        name="somevalue",
        category="somecategory",
        entries=[TextNode("somevalue1"), TextNode("somevalue2")],
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )



from mau.nodes.inline import SentenceNode, TextNode
from mau.nodes.page import (
    BlockNode,
    CommandTocNode,
    ContainerNode,
    ContentImageNode,
    ContentNode,
    DocumentNode,
    HeaderNode,
    HorizontalRuleNode,
    ListNode,
    PageNode,
    ParagraphNode,
)


def test_page_node():
    node = PageNode(
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )

    assert node.args == ["value1", "value2"]
    assert node.tags == ["tag1", "tag2"]
    assert node.kwargs == {"key1": "text1", "key2": "text2"}
    assert node.node_type == "page_node"
    assert node == PageNode(
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )


def test_horizontal_rule_node():
    node = HorizontalRuleNode(
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )

    assert node.args == ["value1", "value2"]
    assert node.tags == ["tag1", "tag2"]
    assert node.kwargs == {"key1": "text1", "key2": "text2"}
    assert node.node_type == "horizontal_rule"
    assert node == HorizontalRuleNode(
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )


def test_paragraph_node():
    node = ParagraphNode(
        SentenceNode([TextNode("sometext")]),
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )

    assert node.content == SentenceNode([TextNode("sometext")])
    assert node.args == ["value1", "value2"]
    assert node.tags == ["tag1", "tag2"]
    assert node.kwargs == {"key1": "text1", "key2": "text2"}
    assert node.node_type == "paragraph"
    assert node == ParagraphNode(
        SentenceNode([TextNode("sometext")]),
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )


def test_command_toc_node():
    node = CommandTocNode(
        entries=[
            HeaderNode(
                "somevalue",
                "somelevel",
                "someanchor",
                args=["hvalue1", "hvalue2"],
                kwargs={"hkey1": "htext1", "hkey2": "htext2"},
            )
        ],
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )

    assert node.entries == [
        HeaderNode(
            "somevalue",
            "somelevel",
            "someanchor",
            args=["hvalue1", "hvalue2"],
            kwargs={"hkey1": "htext1", "hkey2": "htext2"},
        )
    ]
    assert node.args == ["value1", "value2"]
    assert node.tags == ["tag1", "tag2"]
    assert node.kwargs == {"key1": "text1", "key2": "text2"}
    assert node.node_type == "command_toc"
    assert node == CommandTocNode(
        entries=[
            HeaderNode(
                "somevalue",
                "somelevel",
                "someanchor",
                args=["hvalue1", "hvalue2"],
                kwargs={"hkey1": "htext1", "hkey2": "htext2"},
            )
        ],
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )


def test_header_node():
    node = HeaderNode(
        "somevalue",
        "somelevel",
        "someanchor",
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )

    assert node.value == "somevalue"
    assert node.level == "somelevel"
    assert node.anchor == "someanchor"
    assert node.args == ["value1", "value2"]
    assert node.tags == ["tag1", "tag2"]
    assert node.kwargs == {"key1": "text1", "key2": "text2"}
    assert node.node_type == "header"
    assert node == HeaderNode(
        "somevalue",
        "somelevel",
        "someanchor",
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )


def test_list_node():
    node = ListNode(ordered=True, items=[], main_node=False)

    assert node.ordered is True
    assert node.items == []
    assert node.main_node is False
    assert node.args == []
    assert node.kwargs == {}
    assert node.node_type == "list"
    assert node == ListNode(ordered=True, items=[], main_node=False)


def test_content_node():
    node = ContentNode(content_type="somecontent", title="sometitle")

    assert node.content_type == "somecontent"
    assert node.title == "sometitle"
    assert node.args == []
    assert node.kwargs == {}
    assert node.node_type == "content"
    assert node == ContentNode(content_type="somecontent", title="sometitle")


def test_content_image_node():
    node = ContentImageNode(
        uri="someuri", alt_text="somealttext", classes=[], title="sometitle"
    )

    assert node.uri == "someuri"
    assert node.alt_text == "somealttext"
    assert node.classes == []
    assert node.title == "sometitle"
    assert node.args == []
    assert node.kwargs == {}
    assert node.node_type == "content_image"
    assert node == ContentImageNode(
        uri="someuri", alt_text="somealttext", classes=[], title="sometitle"
    )


def test_block_node():
    node = BlockNode(blocktype="sometype", content=[], secondary_content=[])

    assert node.blocktype == "sometype"
    assert node.content == []
    assert node.secondary_content == []
    assert node.args == []
    assert node.kwargs == {}
    assert node.node_type == "block"
    assert node == BlockNode(blocktype="sometype", content=[], secondary_content=[])


def test_container_node():
    node = ContainerNode(content=[])

    assert node.content == []
    assert node.args == []
    assert node.kwargs == {}
    assert node.node_type == "container"
    assert node == ContainerNode(content=[])


def test_document_node():
    node = DocumentNode(content=[])

    assert node.content == []
    assert node.args == []
    assert node.kwargs == {}
    assert node.node_type == "document"
    assert node == DocumentNode(content=[])



from mau.lexers.main_lexer import MainLexer
from mau.nodes.inline import ListItemNode, SentenceNode, TextNode
from mau.nodes.page import ListNode
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_parse_list_with_one_item():
    source = """
    * This is a list with one element
    """

    assert runner(source).nodes == [
        ListNode(
            False,
            [
                ListItemNode(
                    "1", SentenceNode([TextNode("This is a list with one element")])
                )
            ],
            True,
        )
    ]


def test_parse_list_with_multiple_items():
    source = """
    * Item 1
    * Item 2
    """

    assert runner(source).nodes == [
        ListNode(
            False,
            [
                ListItemNode("1", SentenceNode([TextNode("Item 1")])),
                ListItemNode("1", SentenceNode([TextNode("Item 2")])),
            ],
            True,
        )
    ]


def test_parse_list_with_multiple_levels():
    source = """
    * Item 1
    ** Item 1.1
    * Item 2
    """

    assert runner(source).nodes == [
        ListNode(
            False,
            [
                ListItemNode(
                    "1",
                    SentenceNode(
                        [
                            TextNode("Item 1"),
                            ListNode(
                                False,
                                [
                                    ListItemNode(
                                        "2",
                                        SentenceNode(
                                            [
                                                TextNode("Item 1.1"),
                                            ]
                                        ),
                                    )
                                ],
                                False,
                            ),
                        ]
                    ),
                ),
                ListItemNode("1", SentenceNode([TextNode("Item 2")])),
            ],
            True,
        )
    ]


def test_parse_numbered_list():
    source = """
    # Item 1
    ## Item 1.1
    # Item 2
    """

    assert runner(source).nodes == [
        ListNode(
            True,
            [
                ListItemNode(
                    "1",
                    SentenceNode(
                        [
                            TextNode("Item 1"),
                            ListNode(
                                True,
                                [
                                    ListItemNode(
                                        "2",
                                        SentenceNode(
                                            [
                                                TextNode("Item 1.1"),
                                            ]
                                        ),
                                    )
                                ],
                                False,
                            ),
                        ]
                    ),
                ),
                ListItemNode(
                    "1",
                    SentenceNode(
                        [TextNode("Item 2")],
                    ),
                ),
            ],
            True,
        )
    ]


def test_parse_mixed_list():
    source = """
    * Item 1
    ## Item 1.1
    * Item 2
    """

    assert runner(source).nodes == [
        ListNode(
            False,
            [
                ListItemNode(
                    "1",
                    SentenceNode(
                        [
                            TextNode("Item 1"),
                            ListNode(
                                True,
                                [
                                    ListItemNode(
                                        "2",
                                        SentenceNode(
                                            [
                                                TextNode("Item 1.1"),
                                            ]
                                        ),
                                    )
                                ],
                                False,
                            ),
                        ]
                    ),
                ),
                ListItemNode("1", SentenceNode([TextNode("Item 2")])),
            ],
            True,
        )
    ]


def test_parse_mixed_list_cannot_change_type():
    source1 = """
    * Item 1
    ## Item 1.1
    # Item 2
    """

    source2 = """
    * Item 1
    ## Item 1.1
    * Item 2
    """

    assert runner(source1).nodes == runner(source2).nodes



import textwrap
from unittest.mock import patch

from mau.lexers.main_lexer import MainLexer
from mau.nodes.inline import SentenceNode, TextNode
from mau.nodes.page import ParagraphNode
from mau.nodes.references import (
    CommandReferencesNode,
    ReferenceNode,
    ReferencesEntryNode,
)
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


@patch("mau.parsers.text_parser.ReferenceNode")
def test_parse_paragraph_with_reference(mock_reference_node):
    source = """
    This is a paragraph with a [reference](content_type, value)
    """

    parser = runner(source)
    reference_node = mock_reference_node()

    assert parser.nodes == [
        ParagraphNode(
            SentenceNode([TextNode("This is a paragraph with a "), reference_node]),
            args=[],
            kwargs={},
        )
    ]
    assert parser.reference_mentions == {("content_type", "value"): reference_node}
    assert parser.reference_data == {}


def test_parse_document_with_reference():
    source = """
    . Some title
    [reference, content_type, value]
    ----
    This is a paragraph.
    ----
    """

    parser = init_parser(textwrap.dedent(source))
    parser.parse()

    assert parser.nodes == []
    assert parser.reference_data[("content_type", "value")] == {
        "content": [
            ParagraphNode(
                SentenceNode(
                    [
                        TextNode("This is a paragraph."),
                    ]
                )
            ),
        ],
        "title": SentenceNode(
            [
                TextNode("Some title"),
            ]
        ),
    }


@patch("mau.parsers.main_parser.reference_anchor")
def test_parse_reference(mock_reference_anchor):
    source = """
    This is a paragraph with a [reference](content_type, value)

    [reference, content_type, value]
    ----
    This is a paragraph.
    ----
    """

    mock_reference_anchor.return_value = "XXYY"

    parser = runner(source)
    parser.process_references()

    assert parser.references == {
        ("content_type", "value"): ReferenceNode(
            "content_type",
            "value",
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("This is a paragraph."),
                        ]
                    )
                ),
            ],
            number=1,
            reference_anchor="ref-content_type-1-XXYY",
            content_anchor="cnt-content_type-1-XXYY",
        ),
    }


@patch("mau.parsers.main_parser.reference_anchor")
def test_parse_stored_content_multiple_content(mock_reference_anchor):
    mock_reference_anchor.return_value = "XXYY"

    source = """
    [reference, content_type1, value1]
    ----
    Content type 1, value 1
    ----

    [reference, content_type1, value2]
    ----
    Content type 1, value 2
    ----

    [reference, content_type2, value1]
    ----
    Content type 2, value 1
    ----

    This is a paragraph with references of different types:
    [reference](content_type1, value1), [reference](content_type1, value2), [reference](content_type2, value1)
    """

    parser = runner(source)
    parser.process_references()

    assert parser.references == {
        ("content_type1", "value1"): ReferenceNode(
            "content_type1",
            "value1",
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("Content type 1, value 1"),
                        ]
                    )
                ),
            ],
            number=1,
            reference_anchor="ref-content_type1-1-XXYY",
            content_anchor="cnt-content_type1-1-XXYY",
        ),
        ("content_type1", "value2"): ReferenceNode(
            "content_type1",
            "value2",
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("Content type 1, value 2"),
                        ]
                    )
                ),
            ],
            number=2,
            reference_anchor="ref-content_type1-2-XXYY",
            content_anchor="cnt-content_type1-2-XXYY",
        ),
        ("content_type2", "value1"): ReferenceNode(
            "content_type2",
            "value1",
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("Content type 2, value 1"),
                        ]
                    )
                ),
            ],
            number=1,
            reference_anchor="ref-content_type2-1-XXYY",
            content_anchor="cnt-content_type2-1-XXYY",
        ),
    }

    assert parser.reference_entries == {
        ("content_type1", "value1"): ReferencesEntryNode(
            "content_type1",
            "value1",
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("Content type 1, value 1"),
                        ]
                    )
                ),
            ],
            number=1,
            reference_anchor="ref-content_type1-1-XXYY",
            content_anchor="cnt-content_type1-1-XXYY",
        ),
        ("content_type1", "value2"): ReferencesEntryNode(
            "content_type1",
            "value2",
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("Content type 1, value 2"),
                        ]
                    )
                ),
            ],
            number=2,
            reference_anchor="ref-content_type1-2-XXYY",
            content_anchor="cnt-content_type1-2-XXYY",
        ),
        ("content_type2", "value1"): ReferencesEntryNode(
            "content_type2",
            "value1",
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("Content type 2, value 1"),
                        ]
                    )
                ),
            ],
            number=1,
            reference_anchor="ref-content_type2-1-XXYY",
            content_anchor="cnt-content_type2-1-XXYY",
        ),
    }


def test_command_references_parse_args():
    source = "::references:content_type, category, value, arg1, #tag1, name1=value1, name2=value2"

    parser = init_parser(source)
    parser.parse()

    assert parser.nodes == [
        CommandReferencesNode(
            content_type="content_type",
            category="category",
            name="value",
            entries={},
            args=["arg1"],
            kwargs={"name1": "value1", "name2": "value2"},
            tags=["tag1"],
        ),
    ]


@patch("mau.parsers.main_parser.reference_anchor")
def test_command_references_only_content_type(mock_reference_anchor):
    mock_reference_anchor.return_value = "XXYY"

    source = """
    [reference, content_type1, value1]
    ----
    Content type 1, value 1
    ----

    [reference, content_type1, value2]
    ----
    Content type 1, value 2
    ----

    [reference, content_type2, value1]
    ----
    Content type 2, value 1
    ----

    This is a paragraph with references of different types:
    [reference](content_type1, value1), [reference](content_type1, value2), [reference](content_type2, value1)

    ::references:content_type1
    """

    parser = runner(source)
    parser.parse()
    parser.process_references()

    assert parser.nodes == [
        ParagraphNode(
            SentenceNode(
                [
                    TextNode(
                        "This is a paragraph with references of different types: "
                    ),
                    ReferenceNode(
                        "content_type1",
                        "value1",
                        content=[
                            ParagraphNode(
                                SentenceNode([TextNode("Content type 1, value 1")])
                            ),
                        ],
                        number=1,
                        reference_anchor="ref-content_type1-1-XXYY",
                        content_anchor="cnt-content_type1-1-XXYY",
                    ),
                    TextNode(", "),
                    ReferenceNode(
                        "content_type1",
                        "value2",
                        content=[
                            ParagraphNode(
                                SentenceNode([TextNode("Content type 1, value 2")])
                            ),
                        ],
                        number=2,
                        reference_anchor="ref-content_type1-2-XXYY",
                        content_anchor="cnt-content_type1-2-XXYY",
                    ),
                    TextNode(", "),
                    ReferenceNode(
                        "content_type2",
                        "value1",
                        content=[
                            ParagraphNode(
                                SentenceNode([TextNode("Content type 2, value 1")])
                            ),
                        ],
                        number=1,
                        reference_anchor="ref-content_type2-1-XXYY",
                        content_anchor="cnt-content_type2-1-XXYY",
                    ),
                ]
            ),
            args=[],
            kwargs={},
        ),
        CommandReferencesNode(
            content_type="content_type1",
            entries=parser.reference_entries,
            args=[],
            kwargs={},
            tags=[],
        ),
    ]



import yaml

import logging

from mau.parsers.toc import create_toc

_logger = logging.getLogger(__name__)


class VisitorError(ValueError):
    """This is a detailed visitor error"""

    def __init__(self, message, node=None):
        super().__init__(message)

        self.node = node

    def __repr__(self):
        return f"{super().__str__()} - Node: {self.node}"

    def __str__(self):
        return self.__repr__()


class BaseVisitor:
    format_code = "dump"
    extension = ""
    transform = yaml.dump

    # pylint: disable=unused-argument
    def __init__(self, *args, **kwds):
        self.toc = None
        self.footnotes = None

        self._join_with = {}

    def visit(self, node, *args, **kwargs):
        # The visitor has to define functions for each node type.
        # Those shall return a dictionary with the key "data" that
        # contains the result of the visit.
        # This is done to provide space for metadata or other values
        # like templates used to render the node.

        if node is None:
            return {}

        return node.accept(self, *args, **kwargs)

    def visitlist(self, nodes, *args, **kwargs):
        join_with = kwargs.pop("join_with", None)

        visited_nodes = [self.visit(i, *args, **kwargs) for i in nodes]

        if join_with is not None:
            return join_with.join(visited_nodes)

        return visited_nodes

    def _error(self, message, node=None):
        raise VisitorError(message, node)

    def _visit_default(self, node):
        self._error("Cannot find visit function", node)

    # Inline nodes

    def _visit_text(self, node):
        return {
            "data": {
                "type": node.node_type,
                "value": node.value,
            },
        }

    def _visit_verbatim(self, node):
        return {
            "data": {
                "type": node.node_type,
                "value": node.value,
            },
        }

    def _visit_sentence(self, node):
        join_with = self._join_with.get(node.node_type, None)

        return {
            "data": {
                "type": node.node_type,
                "content": self.visitlist(node.content, join_with=join_with),
            },
        }

    def _visit_style(self, node):
        return {
            "data": {
                "type": node.node_type,
                "value": node.value,
                "content": self.visit(node.content),
            },
        }

    def _visit_macro(self, node):
        return {
            "data": {
                "type": node.node_type,
                "name": node.name,
                "args": node.args,
                "kwargs": node.kwargs,
            },
        }

    def _visit_footnote(self, node):
        join_with = self._join_with.get(node.node_type, None)

        return {
            "data": {
                "type": node.node_type,
                "content": self.visitlist(node.content, join_with=join_with),
                "number": node.number,
                "reference_anchor": node.reference_anchor,
                "content_anchor": node.content_anchor,
            },
        }

    def _visit_class(self, node):
        return {
            "data": {
                "type": node.node_type,
                "classes": node.classes,
                "content": self.visit(node.content),
            },
        }

    def _visit_link(self, node):
        return {
            "data": {
                "type": node.node_type,
                "target": node.target,
                "text": node.text,
            },
        }

    def _visit_image(self, node):
        return {
            "data": {
                "type": node.node_type,
                "uri": node.uri,
                "alt_text": node.alt_text,
                "width": node.width,
                "height": node.height,
            },
        }

    def _visit_list_item(self, node):
        return {
            "data": {
                "type": node.node_type,
                "level": int(node.level),
                "content": self.visit(node.content),
            },
        }

    # Page nodes

    def _visit_horizontal_rule(self, node):
        return {
            "data": {
                "type": node.node_type,
                "args": node.args,
                "kwargs": node.kwargs,
                "tags": node.tags,
            },
        }

    def _visit_paragraph(self, node):
        return {
            "data": {
                "type": node.node_type,
                "content": self.visit(node.content),
                "args": node.args,
                "kwargs": node.kwargs,
                "tags": node.tags,
            },
        }

    def _visit_header(self, node):
        return {
            "data": {
                "type": node.node_type,
                "value": node.value,
                "level": int(node.level),
                "anchor": node.anchor,
                "args": node.args,
                "kwargs": node.kwargs,
                "tags": node.tags,
            },
        }

    def _visit_list(self, node):
        join_with = self._join_with.get(node.node_type, None)

        return {
            "data": {
                "type": node.node_type,
                "ordered": node.ordered,
                "items": self.visitlist(node.items, join_with=join_with),
                "main_node": node.main_node,
                "args": node.args,
                "kwargs": node.kwargs,
                "tags": node.tags,
            },
        }

    def _visit_content(self, node):
        return {
            "data": {
                "type": node.node_type,
                "content_type": node.content_type,
                "title": self.visit(node.title),
                "args": node.args,
                "kwargs": node.kwargs,
                "tags": node.tags,
            },
        }

    def _visit_content_image(self, node):
        return {
            "data": {
                "type": node.node_type,
                "uri": node.uri,
                "alt_text": node.alt_text,
                "classes": node.classes,
                "title": self.visit(node.title),
                "args": node.args,
                "kwargs": node.kwargs,
                "tags": node.tags,
            },
        }

    def _visit_block(self, node):
        join_with = self._join_with.get(node.node_type, None)

        return {
            "data": {
                "type": node.node_type,
                "blocktype": node.blocktype,
                "content": self.visitlist(node.content, join_with=join_with),
                "secondary_content": self.visitlist(
                    node.secondary_content, join_with=join_with
                ),
                "classes": node.classes,
                "title": self.visit(node.title),
                "engine": node.engine,
                "preprocessor": node.preprocessor,
                "args": node.args,
                "kwargs": node.kwargs,
                "tags": node.tags,
            }
        }

    def _visit_source(self, node):
        for method_name in [
            "_visit_source__default",
            f"_visit_source_{node.language}",
        ]:
            try:
                method = getattr(self, method_name)
            except AttributeError:
                continue

        return method(node)

    def _visit_source__default(self, node):
        code = self.visitlist(node.code)

        # This is a list of callouts for each line
        # None if the callout is not present, otherwise
        # the rendered callout
        markers = [None] * len(code)

        for marker in node.markers:
            markers[marker.line] = self.visit(marker)

        return {
            "data": {
                "type": node.node_type,
                "blocktype": node.blocktype,
                "code": code,
                "language": node.language,
                "callouts": self.visitlist(node.callouts),
                "highlights": node.highlights,
                "markers": markers,
                "classes": node.classes,
                "title": self.visit(node.title),
                "preprocessor": node.preprocessor,
                "lines": len(node.code),
                "args": node.args,
                "kwargs": node.kwargs,
                "tags": node.tags,
            },
        }

    def _visit_callout(self, node):
        return {
            "data": {
                "type": node.node_type,
                "line": node.line,
                "marker": node.marker,
            },
        }

    def _visit_callouts_entry(self, node):
        return {
            "data": {
                "type": node.node_type,
                "marker": node.marker,
                "value": node.value,
            },
        }

    def _visit_toc_entry(self, node, exclude_tag=None):
        join_with = self._join_with.get(node.node_type, None)

        children = [i for i in node.children if exclude_tag not in i.tags]

        return {
            "data": {
                "type": node.node_type,
                "value": node.value,
                "anchor": node.anchor,
                "children": self.visitlist(children, join_with=join_with),
                "args": self.toc.args,
                "kwargs": self.toc.kwargs,
                "tags": self.toc.tags,
            },
        }

    def _visit_command_toc(self, node):
        join_with = self._join_with.get(node.node_type, None)

        if self.toc is None:
            self.toc = create_toc(
                node.entries, args=node.args, kwargs=node.kwargs, tags=node.tags
            )

        entries = [
            i for i in self.toc.entries if node.kwargs.get("exclude_tag") not in i.tags
        ]

        return {
            "data": {
                "type": node.node_type,
                "entries": self.visitlist(
                    entries,
                    exclude_tag=self.toc.kwargs.get("exclude_tag"),
                    join_with=join_with,
                ),
                "args": self.toc.args,
                "kwargs": self.toc.kwargs,
                "tags": self.toc.tags,
            },
        }

    def _visit_command_footnotes(self, node):
        join_with = self._join_with.get(node.node_type, None)

        return {
            "data": {
                "type": node.node_type,
                "entries": self.visitlist(node.entries, join_with=join_with),
                "args": node.args,
                "kwargs": node.kwargs,
                "tags": node.tags,
            },
        }

    def _visit_footnotes_entry(self, node):
        join_with = self._join_with.get(node.node_type, None)

        return {
            "data": {
                "type": node.node_type,
                "content": self.visitlist(node.content, join_with=join_with),
                "number": node.number,
                "reference_anchor": node.reference_anchor,
                "content_anchor": node.content_anchor,
            },
        }

    def _visit_container(self, node):
        join_with = self._join_with.get(node.node_type, None)

        return {
            "data": {
                "type": node.node_type,
                "content": self.visitlist(node.content, join_with=join_with),
                "args": node.args,
                "kwargs": node.kwargs,
                "tags": node.tags,
            },
        }

    def _visit_document(self, node):
        return self._visit_container(node)

    def _visit_reference(self, node):
        join_with = self._join_with.get(node.node_type, None)

        return {
            "data": {
                "type": node.node_type,
                "content_type": node.content_type,
                "name": node.name,
                "category": node.category,
                "content": self.visitlist(node.content, join_with=join_with),
                "number": node.number,
                "title": self.visit(node.title),
                "reference_anchor": node.reference_anchor,
                "content_anchor": node.content_anchor,
            },
        }

    def _visit_command_references(self, node):
        join_with = self._join_with.get(node.node_type, None)

        entries = [
            i for i in node.entries.values() if i.content_type == node.content_type
        ]

        if node.category is not None:
            entries = [i for i in entries if i.category == node.category]

        if node.name is not None:
            entries = [i for i in entries if i.name == node.name]

        return {
            "data": {
                "type": node.node_type,
                "content_type": node.content_type,
                "category": node.category,
                "name": node.name,
                "entries": self.visitlist(entries, join_with=join_with),
                "args": node.args,
                "kwargs": node.kwargs,
                "tags": node.tags,
            },
        }

    def _visit_references_entry(self, node):
        join_with = self._join_with.get(node.node_type, None)

        return {
            "data": {
                "type": node.node_type,
                "content_type": node.content_type,
                "name": node.name,
                "category": node.category,
                "content": self.visitlist(node.content, join_with=join_with),
                "number": node.number,
                "title": self.visit(node.title),
                "reference_anchor": node.reference_anchor,
                "content_anchor": node.content_anchor,
            },
        }



from mau.helpers import rematch
from mau.tokens.tokens import Token


class LexerError(ValueError):
    """This is a detailed lexer error"""

    def __init__(self, context, message=None):
        super().__init__(message)

        self.context = context


class TokenTypes:
    EMPTY = "EMPTY"
    EOF = "EOF"
    EOL = "EOL"
    LITERAL = "LITERAL"
    TEXT = "TEXT"
    WHITESPACE = "WHITESPACE"


class BaseLexer:
    """
    The base class for lexers.
    The lexer decomposes the input text into a list of tokens
    and provides basic navigation functions in the
    output results.
    """

    def __init__(self, text_buffer):
        self._text_buffer = text_buffer

        # These are the tokens identified so far
        self.tokens = []

        # The last visited context. Used to detect loops.
        self.last_visited_context = None

    def process(self):
        # Process tokens until we reach the end of file
        self._process()
        while True:
            # Check if the last thing we processed is an EOF
            if len(self.tokens) > 0 and self.tokens[-1].type is TokenTypes.EOF:
                break

            self._process()

        if Token(TokenTypes.EOF) in self.tokens:
            self.tokens.remove(Token(TokenTypes.EOF))

    @property
    def _current_char(self):
        # Return the current character
        return self._text_buffer.current_char

    @property
    def _current_line(self):
        # Return the current line
        return self._text_buffer.current_line

    @property
    def _context(self):
        # Return the context
        return self._text_buffer.context

    @property
    def _tail(self):
        # A wrapper to return the rest of the line
        return self._text_buffer.tail

    def _nextline(self):
        # Skip the whole line including the EOL
        self._text_buffer.nextline()

    def _skip(self, value):
        # Skip only the given amount of characters
        # This is very useful with regexp groups
        # that can be None.
        if value is not None:
            self._text_buffer.skip(len(value))

    def _error(self, message=None):
        raise LexerError(context=self._context, message=message)

    def _process(self):
        # This should not be touched by child classes
        # as it is the core of the lexer. It tries
        # each function in the list returned by
        # _process_functions and stores all the resulting
        # tokens.
        # All lexers process first EOF and EOL, and last
        # an error. This is mandatory as the underlying
        # text buffer doesn't flinch when we are past
        # EOL or EOF, and returns an empty string.
        # However, this means the parse can't skip the
        # token (it's empty) and end up in an infinite
        # loop, so we have to actively check that.
        # A lexer function must return None
        # when characters do not match the rule.

        process_functions = [self._process_eof, self._process_eol]
        process_functions.extend(self._process_functions())
        process_functions.append(self._process_error)

        # This detects infinite loops created by incomplete
        # lexing functions. Those functions keep trying
        # to parse the same context, so if we spot that
        # we are doing it we should raise an error.
        if (
            self.last_visited_context is not None
            and self.last_visited_context == self._context
        ):
            self._error("Loop detected, cannot process context.")  # pragma: no cover

        self.last_visited_context = self._context

        for process_func in process_functions:
            # This ensures result is always either None or a list
            result = process_func()

            if result is None:
                continue

            self.tokens.extend(result)

            return

    def _process_functions(self):
        return [
            self._process_text_line,
        ]

    def _process_error(self):
        self._error("No function found to process context.")

    def _create_token(self, token_type, token_value=None):
        # Create a token at the current position,
        # passing the current context.
        # This doesn't change the position in the buffer

        return Token(token_type, token_value, self._context)

    def _create_token_and_skip(self, token_type, token_value):
        # Create the token and advance the position
        # in the text buffer to skip the characters
        # that are part of the token.

        token = self._create_token(token_type, token_value)

        self._skip(token_value)

        return token

    def _create_tokens_from_line(self, token_type):
        # Create a token of the given type using the
        # whole current line as value.
        # Returns the token and EOL, then moves
        # to the next line.

        tokens = [
            self._create_token_and_skip(token_type, self._current_line),
            self._create_token(TokenTypes.EOL),
        ]

        self._nextline()

        return tokens

    def _process_eof(self):
        if self._text_buffer.eof:
            return [self._create_token(TokenTypes.EOF)]

        return None

    def _process_eol(self):
        # This cannot use TextBuffer.eol as we want to eat any spaces
        # that are present in the line.
        match = rematch(r"\ *$", self._tail)

        if not match:
            return None

        token = self._create_token_and_skip(TokenTypes.EOL, self._tail)

        self._nextline()

        return [token]

    def _process_text_line(self):
        tokens = [
            self._create_token_and_skip(TokenTypes.TEXT, self._tail),
            self._create_token(TokenTypes.EOL),
        ]

        self._nextline()

        return tokens



from unittest.mock import Mock

import pytest
from mau.lexers.base_lexer import BaseLexer, TokenTypes
from mau.parsers.base_parser import BaseParser, ParserError, TokenError
from mau.tokens.tokens import Token

from tests.helpers import init_parser_factory

init_parser = init_parser_factory(BaseLexer, BaseParser)


def test_save():
    parser = BaseParser(tokens=[])
    node = Mock()
    parser._save(node)

    assert parser.nodes == [node]


def test_initial_state():
    parser = BaseParser(tokens=[])

    assert parser.index == -1
    assert parser._current_token == Token(TokenTypes.EOF)


def test_load():
    parser = init_parser("\n")

    assert parser.index == -1
    assert parser._current_token == Token(TokenTypes.EOF)


def test_get_token_without_load():
    parser = BaseParser(tokens=[])

    assert parser._get_token() == Token(TokenTypes.EOF)


def test_get_token():
    parser = init_parser("")

    assert parser._get_token() == Token(TokenTypes.EOF)


def test_put_token():
    parser = init_parser("")

    parser._put_token(Token(TokenTypes.LITERAL, "*"))
    assert parser._get_token() == Token(TokenTypes.LITERAL, "*")
    assert parser._get_token() == Token(TokenTypes.EOF)


def test_get_token_after_eof():
    parser = init_parser("")

    assert parser.index == -1

    assert parser._get_token() == Token(TokenTypes.EOF)
    assert parser.index == 0

    assert parser._get_token() == Token(TokenTypes.EOF)
    assert parser.index == 0


def test_get_token_sets_current_token():
    parser = init_parser("Some text\nSome other text")

    assert parser._get_token() == Token(TokenTypes.TEXT, "Some text")
    assert parser._current_token == Token(TokenTypes.TEXT, "Some text")

    assert parser._get_token() == Token(TokenTypes.EOL)
    assert parser._current_token == Token(TokenTypes.EOL)

    assert parser._get_token() == Token(TokenTypes.TEXT, "Some other text")
    assert parser._current_token == Token(TokenTypes.TEXT, "Some other text")


def test_get_token_accepts_type():
    parser = init_parser("Some text\nSome other text")

    assert parser._get_token(TokenTypes.TEXT) == Token(TokenTypes.TEXT, "Some text")


def test_get_token_accepts_type_and_value():
    parser = init_parser("Some text\nSome other text")

    assert parser._get_token(TokenTypes.TEXT, "Some text") == Token(
        TokenTypes.TEXT, "Some text"
    )


def test_get_token_checks_type():
    parser = init_parser("Some text\nSome other text")

    with pytest.raises(TokenError):
        parser._get_token(TokenTypes.EOL)


def test_get_token_checks_type_and_value():
    parser = init_parser("Some text\nSome other text")

    with pytest.raises(TokenError):
        parser._get_token(TokenTypes.TEXT, "foobar")


def test_get_token_accepts_check_function():
    parser = init_parser("Some text\nSome other text")

    with pytest.raises(TokenError):
        parser._get_token(TokenTypes.TEXT, check=lambda x: x == "foobar")


def test_check_current_token():
    parser = init_parser("Some text\nSome other text")
    parser._get_token()

    assert parser._check_current_token(TokenTypes.TEXT) == Token(
        TokenTypes.TEXT, "Some text"
    )

    # check_current_token doesn't advance the index
    assert parser._get_token() == Token(TokenTypes.EOL)


def test_check_current_token_raises_exception():
    parser = init_parser("Some text\nSome other text")
    parser._get_token()

    with pytest.raises(TokenError):
        parser._check_current_token(TokenTypes.EOF)


def test_check_current_token_value():
    parser = init_parser("Some text\nSome other text")
    parser._get_token()

    assert parser._check_current_token(TokenTypes.TEXT, "Some text") == Token(
        TokenTypes.TEXT, "Some text"
    )


def test_check_current_token_with_function():
    parser = init_parser("Some text\nSome other text")
    parser._get_token()

    assert parser._check_current_token(
        TokenTypes.TEXT, check=lambda x: x == "Some text"
    ) == Token(TokenTypes.TEXT, "Some text")


def test_peek_token():
    parser = init_parser("Some text\nSome other text")

    assert parser._peek_token() == Token(TokenTypes.TEXT, "Some text")
    assert parser._get_token() == Token(TokenTypes.TEXT, "Some text")
    assert parser._get_token() == Token(TokenTypes.EOL)
    assert parser._get_token() == Token(TokenTypes.TEXT, "Some other text")
    assert parser._peek_token() == Token(TokenTypes.EOL)
    assert parser._get_token() == Token(TokenTypes.EOL)
    assert parser._peek_token() == Token(TokenTypes.EOF)
    assert parser._get_token() == Token(TokenTypes.EOF)


def test_base_lexer_can_peek_after_eof():
    parser = init_parser("Some text\nSome other text")

    assert parser._get_token() == Token(TokenTypes.TEXT, "Some text")
    assert parser._get_token() == Token(TokenTypes.EOL)
    assert parser._get_token() == Token(TokenTypes.TEXT, "Some other text")
    assert parser._get_token() == Token(TokenTypes.EOL)
    assert parser._get_token() == Token(TokenTypes.EOF)
    assert parser._peek_token() == Token(TokenTypes.EOF)


def test_peek_token_checks_type():
    parser = init_parser("Some text\nSome other text")

    assert parser._peek_token(TokenTypes.TEXT) == Token(TokenTypes.TEXT, "Some text")


def test_base_lexer_check_next_token_raises_exception():
    parser = init_parser("Some text\nSome other text")

    with pytest.raises(TokenError):
        parser._peek_token(TokenTypes.EOL)

    # failed attempt doesn't advance the index
    assert parser._get_token() == Token(TokenTypes.TEXT, "Some text")


def test_peek_token_checks_type_and_value():
    parser = init_parser("Some text\nSome other text")

    with pytest.raises(TokenError):
        parser._peek_token(TokenTypes.EOL, "foobar")


def test_peek_token_accepts_check_function():
    parser = init_parser("Some text\nSome other text")

    parser._peek_token(TokenTypes.TEXT, check=lambda x: x == "Some text")

    with pytest.raises(TokenError):
        parser._peek_token(TokenTypes.TEXT, check=lambda x: x == "Some other text")


def test_context_manager():
    parser = init_parser("Some text\nSome other text")

    with parser:
        assert parser._get_token() == Token(TokenTypes.TEXT, "Some text")


def test_do_not_restore_status_if_no_error():
    parser = init_parser("Some text\nSome other text")

    with parser:
        assert parser._get_token(TokenTypes.TEXT, "Some text")
        assert parser._get_token(TokenTypes.EOL)
        assert parser._get_token(TokenTypes.TEXT, "Some other text")
        assert parser._get_token(TokenTypes.EOL)

    assert parser._get_token(TokenTypes.EOF)


def test_restore_status_if_token_error():
    parser = init_parser("Some text\nSome other text")

    with parser:
        assert parser._get_token(TokenTypes.TEXT, "Some text")
        assert parser._get_token(TokenTypes.TEXT, "Some other text")
        assert parser._get_token(TokenTypes.TEXT, "Some final text")

    assert parser._get_token(TokenTypes.TEXT, "Some text")


def test_context_manager_leaves_exceptions_untouched():
    parser = init_parser("Some text\nSome other text")

    with pytest.raises(ValueError):
        with parser:
            raise ValueError


def test_token_error_can_be_raised():
    parser = init_parser("Some text\nSome other text")

    with parser:
        assert parser._get_token(TokenTypes.TEXT, "Some text")
        raise TokenError

    assert parser._get_token(TokenTypes.TEXT, "Some text")


def test_context_manager_nested_success():
    parser = init_parser("Some text\nSome other text")

    with parser:
        parser._get_token(TokenTypes.TEXT, "Some text")
        parser._get_token(TokenTypes.EOL)
        with parser:
            parser._get_token(TokenTypes.TEXT, "Some other text")
            parser._get_token(TokenTypes.EOL)

        parser._get_token(TokenTypes.EOF)

    assert parser._get_token(TokenTypes.EOF)


def test_context_manager_nested_failure():
    parser = init_parser("Some text\nSome other text")

    with parser:
        parser._get_token(TokenTypes.TEXT, "Some text")
        parser._get_token(TokenTypes.EOL)
        with parser:
            parser._get_token(TokenTypes.TEXT, "Some text")

        parser._get_token(TokenTypes.TEXT, "Some other text")
        parser._get_token(TokenTypes.EOL)

    assert parser._get_token(TokenTypes.EOF)


def test_peek_token_is():
    parser = init_parser("Some text\nSome other text")

    assert parser._peek_token_is(TokenTypes.EOL) is False
    assert parser._peek_token_is(TokenTypes.TEXT) is True


def test_force_token():
    parser = init_parser("Some text\nSome other text")

    assert parser._force_token(TokenTypes.TEXT) == Token(TokenTypes.TEXT, "Some text")
    assert parser._force_token(TokenTypes.EOL) == Token(TokenTypes.EOL)
    assert parser._force_token(TokenTypes.TEXT) == Token(
        TokenTypes.TEXT, "Some other text"
    )

    with pytest.raises(ParserError):
        parser._force_token(TokenTypes.TEXT)


def test_collect():
    parser = init_parser("Some text\nSome other text")

    tokens = parser._collect([Token(TokenTypes.EOF)])

    assert tokens == [
        Token(TokenTypes.TEXT, "Some text"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.TEXT, "Some other text"),
        Token(TokenTypes.EOL),
    ]

    tokens = parser._collect([Token(TokenTypes.EOF)])

    assert tokens == []


def test_collect_join():
    parser = init_parser("Some te\nxt that will be joined\n!")

    expected = "Some text that will be joined!"

    assert parser._collect_join([Token(TokenTypes.EOF)]) == expected


def test_collect_join_with_different_joiner():
    parser = init_parser("Some te\nxt that will be joined\n!")

    expected = "Some te-xt that will be joined-!"

    assert parser._collect_join([Token(TokenTypes.EOF)], "-") == expected


def test_collect_escapes_are_kept():
    parser = init_parser("")
    parser.tokens = [
        Token(TokenTypes.TEXT, "Some text"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.TEXT, "Some other text"),
    ]

    tokens = parser._collect([Token(TokenTypes.EOF)])

    assert tokens == [
        Token(TokenTypes.TEXT, "Some text"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.TEXT, "Some other text"),
    ]


def test_collect_escape_stop_tokens_are_removed():
    parser = init_parser("")
    parser.tokens = [
        Token(TokenTypes.TEXT, "Some text"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "["),
    ]

    tokens = parser._collect([Token(TokenTypes.LITERAL, "[")])

    assert tokens == [
        Token(TokenTypes.TEXT, "Some text"),
        Token(TokenTypes.LITERAL, "["),
    ]


def test_collect_escape_stop_tokens_are_removed2():
    parser = init_parser("")
    parser.tokens = [
        Token(TokenTypes.TEXT, "Some text"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "["),
    ]

    tokens = parser._collect(
        [Token(TokenTypes.LITERAL, "[")], preserve_escaped_stop_tokens=True
    )

    assert tokens == [
        Token(TokenTypes.TEXT, "Some text"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "["),
    ]


def test_error():
    parser = init_parser("Some text")
    parser._get_token()
    parser._get_token()

    with pytest.raises(ParserError) as exception:
        parser._error()

    assert exception.value.token == Token(TokenTypes.EOF)


def test_error_with_message():
    parser = init_parser("Some text")
    parser._get_token()
    parser._get_token()

    with pytest.raises(ParserError) as exception:
        parser._error("Some message")

    assert exception.value.args == ("Some message",)
    assert exception.value.token == Token(TokenTypes.EOF)


def test_unknown_token():
    unknown = Token("UNKNOWN")
    parser = BaseParser(tokens=[unknown])

    with pytest.raises(ParserError) as exception:
        parser.parse()

    assert exception.value.token == unknown



from unittest.mock import mock_open, patch

from mau.lexers.base_lexer import TokenTypes as BLTokenTypes
from mau.lexers.main_lexer import MainLexer, TokenTypes
from mau.text_buffer.context import Context
from mau.text_buffer.text_buffer import TextBuffer
from mau.tokens.tokens import Token

from tests.helpers import dedent


def test_empty_text():
    text_buffer = TextBuffer("")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == []


def test_empty_lines():
    text_buffer = TextBuffer("\n")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [Token(BLTokenTypes.EOL), Token(BLTokenTypes.EOL)]


def test_lines_with_only_spaces():
    text_buffer = TextBuffer("      \n      ")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [Token(BLTokenTypes.EOL), Token(BLTokenTypes.EOL)]


def test_horizontal_rule():
    text_buffer = TextBuffer("---")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.HORIZONTAL_RULE, "---"),
        Token(BLTokenTypes.EOL),
    ]


def test_escape_line():
    text_buffer = TextBuffer(r"\[name]")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(BLTokenTypes.TEXT, r"\[name]"),
        Token(BLTokenTypes.EOL),
    ]


def test_escape_line_beginning_with_backslash():
    text_buffer = TextBuffer(r"\\[name]")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(BLTokenTypes.TEXT, r"\\[name]"),
        Token(BLTokenTypes.EOL),
    ]


def test_arguments():
    text_buffer = TextBuffer("[name]")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.ARGUMENTS, "["),
        Token(BLTokenTypes.TEXT, "name"),
        Token(BLTokenTypes.LITERAL, "]"),
        Token(BLTokenTypes.EOL),
    ]


def test_attributes_no_closing_bracket():
    text_buffer = TextBuffer("[name")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(BLTokenTypes.TEXT, "[name"),
        Token(BLTokenTypes.EOL),
    ]


def test_attributes_marker_in_text():
    text_buffer = TextBuffer("Not [attributes]")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(BLTokenTypes.TEXT, "Not [attributes]"),
        Token(BLTokenTypes.EOL),
    ]

    assert [i.context for i in lex.tokens] == [
        Context(0, 0, None, "Not [attributes]"),
        Context(0, 16, None, "Not [attributes]"),
    ]


def test_variable_definition():
    text_buffer = TextBuffer(":variable:value123")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.VARIABLE, ":"),
        Token(BLTokenTypes.TEXT, "variable"),
        Token(BLTokenTypes.LITERAL, ":"),
        Token(BLTokenTypes.TEXT, "value123"),
        Token(BLTokenTypes.EOL),
    ]

    assert [i.context for i in lex.tokens] == [
        Context(0, 0, None, ":variable:value123"),
        Context(0, 1, None, ":variable:value123"),
        Context(0, 9, None, ":variable:value123"),
        Context(0, 10, None, ":variable:value123"),
        Context(0, 18, None, ":variable:value123"),
    ]


def test_variable_negation():
    text_buffer = TextBuffer(":!variable:")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.VARIABLE, ":"),
        Token(BLTokenTypes.TEXT, "!variable"),
        Token(BLTokenTypes.LITERAL, ":"),
        Token(BLTokenTypes.EOL),
    ]


def test_variable_marker_in_text():
    text_buffer = TextBuffer("Not a :variable:")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(BLTokenTypes.TEXT, "Not a :variable:"),
        Token(BLTokenTypes.EOL),
    ]


def test_variable_definition_accepted_characters():
    text_buffer = TextBuffer(":abcAB.C0123-_:value123")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.VARIABLE, ":"),
        Token(BLTokenTypes.TEXT, "abcAB.C0123-_"),
        Token(BLTokenTypes.LITERAL, ":"),
        Token(BLTokenTypes.TEXT, "value123"),
        Token(BLTokenTypes.EOL),
    ]


def test_multiple_lines():
    text_buffer = TextBuffer(
        dedent(
            """
            This is text
            split into multiple lines

            with an empty line
            """
        )
    )
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(BLTokenTypes.TEXT, "This is text"),
        Token(BLTokenTypes.EOL),
        Token(BLTokenTypes.TEXT, "split into multiple lines"),
        Token(BLTokenTypes.EOL),
        Token(BLTokenTypes.EOL),
        Token(BLTokenTypes.TEXT, "with an empty line"),
        Token(BLTokenTypes.EOL),
    ]

    assert [i.context for i in lex.tokens] == [
        Context(0, 0, None, "This is text"),
        Context(0, 12, None, "This is text"),
        Context(1, 0, None, "split into multiple lines"),
        Context(1, 25, None, "split into multiple lines"),
        Context(2, 0, None, ""),
        Context(3, 0, None, "with an empty line"),
        Context(3, 18, None, "with an empty line"),
    ]


def test_title():
    text_buffer = TextBuffer(".A title")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.TITLE, "."),
        Token(BLTokenTypes.TEXT, "A title"),
        Token(BLTokenTypes.EOL),
    ]

    assert [i.context for i in lex.tokens] == [
        Context(0, 0, None, ".A title"),
        Context(0, 1, None, ".A title"),
        Context(0, 8, None, ".A title"),
    ]


def test_title_with_space():
    text_buffer = TextBuffer(".      A title")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.TITLE, "."),
        Token(BLTokenTypes.WHITESPACE, "      "),
        Token(BLTokenTypes.TEXT, "A title"),
        Token(BLTokenTypes.EOL),
    ]

    assert [i.context for i in lex.tokens] == [
        Context(0, 0, None, ".      A title"),
        Context(0, 1, None, ".      A title"),
        Context(0, 7, None, ".      A title"),
        Context(0, 14, None, ".      A title"),
    ]


def test_command():
    text_buffer = TextBuffer("::command:arg0,arg1")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.COMMAND, "::"),
        Token(BLTokenTypes.TEXT, "command"),
        Token(BLTokenTypes.LITERAL, ":"),
        Token(BLTokenTypes.TEXT, "arg0,arg1"),
        Token(BLTokenTypes.EOL),
    ]

    assert [i.context for i in lex.tokens] == [
        Context(0, 0, None, "::command:arg0,arg1"),
        Context(0, 2, None, "::command:arg0,arg1"),
        Context(0, 9, None, "::command:arg0,arg1"),
        Context(0, 10, None, "::command:arg0,arg1"),
        Context(0, 19, None, "::command:arg0,arg1"),
    ]


def test_command_without_arguments():
    text_buffer = TextBuffer("::command:")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.COMMAND, "::"),
        Token(BLTokenTypes.TEXT, "command"),
        Token(BLTokenTypes.LITERAL, ":"),
        Token(BLTokenTypes.EOL),
    ]


def test_comment():
    text_buffer = TextBuffer("// Some comment")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.COMMENT, "// Some comment"),
        Token(BLTokenTypes.EOL),
    ]


def test_multiline_comment():
    text_buffer = TextBuffer(
        dedent(
            """
            ////
            Some comment

               another line
            ////
            """
        )
    )
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.MULTILINE_COMMENT, "////"),
        Token(BLTokenTypes.EOL),
        Token(BLTokenTypes.TEXT, "Some comment"),
        Token(BLTokenTypes.EOL),
        Token(BLTokenTypes.EOL),
        Token(BLTokenTypes.TEXT, "   another line"),
        Token(BLTokenTypes.EOL),
        Token(TokenTypes.MULTILINE_COMMENT, "////"),
        Token(BLTokenTypes.EOL),
    ]

    assert [i.context for i in lex.tokens] == [
        Context(0, 0, None, "////"),
        Context(0, 4, None, "////"),
        Context(1, 0, None, "Some comment"),
        Context(1, 12, None, "Some comment"),
        Context(2, 0, None, ""),
        Context(3, 0, None, "   another line"),
        Context(3, 15, None, "   another line"),
        Context(4, 0, None, "////"),
        Context(4, 4, None, "////"),
    ]


def test_include_content():
    text_buffer = TextBuffer("<<type:/path/to/it.jpg")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.INCLUDE, "<<"),
        Token(BLTokenTypes.TEXT, "type"),
        Token(BLTokenTypes.LITERAL, ":"),
        Token(BLTokenTypes.TEXT, "/path/to/it.jpg"),
        Token(BLTokenTypes.EOL),
    ]

    assert [i.context for i in lex.tokens] == [
        Context(0, 0, None, "<<type:/path/to/it.jpg"),
        Context(0, 2, None, "<<type:/path/to/it.jpg"),
        Context(0, 6, None, "<<type:/path/to/it.jpg"),
        Context(0, 7, None, "<<type:/path/to/it.jpg"),
        Context(0, 22, None, "<<type:/path/to/it.jpg"),
    ]


def test_include_content_with_space():
    text_buffer = TextBuffer("<<      type:/path/to/it.jpg")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.INCLUDE, "<<"),
        Token(BLTokenTypes.WHITESPACE, "      "),
        Token(BLTokenTypes.TEXT, "type"),
        Token(BLTokenTypes.LITERAL, ":"),
        Token(BLTokenTypes.TEXT, "/path/to/it.jpg"),
        Token(BLTokenTypes.EOL),
    ]

    assert [i.context for i in lex.tokens] == [
        Context(0, 0, None, "<<      type:/path/to/it.jpg"),
        Context(0, 2, None, "<<      type:/path/to/it.jpg"),
        Context(0, 8, None, "<<      type:/path/to/it.jpg"),
        Context(0, 12, None, "<<      type:/path/to/it.jpg"),
        Context(0, 13, None, "<<      type:/path/to/it.jpg"),
        Context(0, 28, None, "<<      type:/path/to/it.jpg"),
    ]


def test_include_content_without_arguments():
    text_buffer = TextBuffer("<<type:")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.INCLUDE, "<<"),
        Token(BLTokenTypes.TEXT, "type"),
        Token(BLTokenTypes.LITERAL, ":"),
        Token(BLTokenTypes.EOL),
    ]

    assert [i.context for i in lex.tokens] == [
        Context(0, 0, None, "<<type:"),
        Context(0, 2, None, "<<type:"),
        Context(0, 6, None, "<<type:"),
        Context(0, 7, None, "<<type:"),
    ]


def test_unordered_list():
    text_buffer = TextBuffer("* Item")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LIST, "*"),
        Token(BLTokenTypes.WHITESPACE, " "),
        Token(BLTokenTypes.TEXT, "Item"),
        Token(BLTokenTypes.EOL),
    ]

    assert [i.context for i in lex.tokens] == [
        Context(0, 0, None, "* Item"),
        Context(0, 1, None, "* Item"),
        Context(0, 2, None, "* Item"),
        Context(0, 6, None, "* Item"),
    ]


def test_unordered_list_leading_space():
    text_buffer = TextBuffer("    * Item")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(BLTokenTypes.WHITESPACE, "    "),
        Token(TokenTypes.LIST, "*"),
        Token(BLTokenTypes.WHITESPACE, " "),
        Token(BLTokenTypes.TEXT, "Item"),
        Token(BLTokenTypes.EOL),
    ]


def test_unordered_list_trailing_space():
    text_buffer = TextBuffer("*       Item")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LIST, "*"),
        Token(BLTokenTypes.WHITESPACE, "       "),
        Token(BLTokenTypes.TEXT, "Item"),
        Token(BLTokenTypes.EOL),
    ]


def test_unordered_list_multiple_stars():
    text_buffer = TextBuffer("*** Item")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LIST, "***"),
        Token(BLTokenTypes.WHITESPACE, " "),
        Token(BLTokenTypes.TEXT, "Item"),
        Token(BLTokenTypes.EOL),
    ]


def test_ordered_list():
    text_buffer = TextBuffer("# Item")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LIST, "#"),
        Token(BLTokenTypes.WHITESPACE, " "),
        Token(BLTokenTypes.TEXT, "Item"),
        Token(BLTokenTypes.EOL),
    ]


def test_ordered_list_multiple_stars():
    text_buffer = TextBuffer("### Item")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.LIST, "###"),
        Token(BLTokenTypes.WHITESPACE, " "),
        Token(BLTokenTypes.TEXT, "Item"),
        Token(BLTokenTypes.EOL),
    ]


def test_header():
    text_buffer = TextBuffer("=Header")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.HEADER, "="),
        Token(BLTokenTypes.TEXT, "Header"),
        Token(BLTokenTypes.EOL),
    ]

    assert [i.context for i in lex.tokens] == [
        Context(0, 0, None, "=Header"),
        Context(0, 1, None, "=Header"),
        Context(0, 7, None, "=Header"),
    ]


def test_header_with_space():
    text_buffer = TextBuffer("=    Header")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.HEADER, "="),
        Token(BLTokenTypes.WHITESPACE, "    "),
        Token(BLTokenTypes.TEXT, "Header"),
        Token(BLTokenTypes.EOL),
    ]


def test_empty_header():
    text_buffer = TextBuffer("=")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(BLTokenTypes.TEXT, "="),
        Token(BLTokenTypes.EOL),
    ]


def test_multiple_header_markers():
    text_buffer = TextBuffer("=== Header")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.HEADER, "==="),
        Token(BLTokenTypes.WHITESPACE, " "),
        Token(BLTokenTypes.TEXT, "Header"),
        Token(BLTokenTypes.EOL),
    ]


def test_header_marker_in_header_text():
    text_buffer = TextBuffer("= a=b")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.HEADER, "="),
        Token(BLTokenTypes.WHITESPACE, " "),
        Token(BLTokenTypes.TEXT, "a=b"),
        Token(BLTokenTypes.EOL),
    ]


def test_header_markers_in_text():
    text_buffer = TextBuffer("Definitely not a === header")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(BLTokenTypes.TEXT, "Definitely not a === header"),
        Token(BLTokenTypes.EOL),
    ]


@patch("mau.lexers.main_lexer.MainLexer._run_directive")
def test_directive(mock_run_directive):
    text_buffer = TextBuffer("::#name:/path/to/file")
    lex = MainLexer(text_buffer)
    lex.process()

    assert mock_run_directive.called_with("name", "/path/to/file")


@patch("builtins.open", new_callable=mock_open, read_data="just some data")
def test_import_directive(mock_file):  # pylint: disable=unused-argument
    text_buffer = TextBuffer("::#include:/path/to/file")
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(BLTokenTypes.TEXT, "just some data"),
        Token(BLTokenTypes.EOL),
    ]

    assert [i.context for i in lex.tokens] == [
        Context(0, 0, "/path/to/file", "just some data"),
        Context(0, 14, "/path/to/file", "just some data"),
    ]


def test_block():
    text_buffer = TextBuffer(
        dedent(
            """
            ----
            Some comment

               another line
            ----
            """
        )
    )
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.BLOCK, "----"),
        Token(BLTokenTypes.EOL),
        Token(BLTokenTypes.TEXT, "Some comment"),
        Token(BLTokenTypes.EOL),
        Token(BLTokenTypes.EOL),
        Token(BLTokenTypes.TEXT, "   another line"),
        Token(BLTokenTypes.EOL),
        Token(TokenTypes.BLOCK, "----"),
        Token(BLTokenTypes.EOL),
    ]

    assert [i.context for i in lex.tokens] == [
        Context(0, 0, None, "----"),
        Context(0, 4, None, "----"),
        Context(1, 0, None, "Some comment"),
        Context(1, 12, None, "Some comment"),
        Context(2, 0, None, ""),
        Context(3, 0, None, "   another line"),
        Context(3, 15, None, "   another line"),
        Context(4, 0, None, "----"),
        Context(4, 4, None, "----"),
    ]


def test_block_four_characters():
    text_buffer = TextBuffer(
        dedent(
            """
            ####
            Some comment

               another line
            ####
            """
        )
    )
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.BLOCK, "####"),
        Token(BLTokenTypes.EOL),
        Token(BLTokenTypes.TEXT, "Some comment"),
        Token(BLTokenTypes.EOL),
        Token(BLTokenTypes.EOL),
        Token(BLTokenTypes.TEXT, "   another line"),
        Token(BLTokenTypes.EOL),
        Token(TokenTypes.BLOCK, "####"),
        Token(BLTokenTypes.EOL),
    ]

    assert [i.context for i in lex.tokens] == [
        Context(0, 0, None, "####"),
        Context(0, 4, None, "####"),
        Context(1, 0, None, "Some comment"),
        Context(1, 12, None, "Some comment"),
        Context(2, 0, None, ""),
        Context(3, 0, None, "   another line"),
        Context(3, 15, None, "   another line"),
        Context(4, 0, None, "####"),
        Context(4, 4, None, "####"),
    ]


def test_block_with_comment():
    text_buffer = TextBuffer(
        dedent(
            """
            ----
            // Comment
            ----
            """
        )
    )
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(TokenTypes.BLOCK, "----"),
        Token(BLTokenTypes.EOL),
        Token(TokenTypes.COMMENT, "// Comment"),
        Token(BLTokenTypes.EOL),
        Token(TokenTypes.BLOCK, "----"),
        Token(BLTokenTypes.EOL),
    ]

    assert [i.context for i in lex.tokens] == [
        Context(0, 0, None, "----"),
        Context(0, 4, None, "----"),
        Context(1, 0, None, "// Comment"),
        Context(1, 10, None, "// Comment"),
        Context(2, 0, None, "----"),
        Context(2, 4, None, "----"),
    ]


def test_block_has_to_begin_with_four_identical_characters():
    text_buffer = TextBuffer(
        dedent(
            """
            abcd
            """
        )
    )
    lex = MainLexer(text_buffer)
    lex.process()

    assert lex.tokens == [
        Token(BLTokenTypes.TEXT, "abcd"),
        Token(BLTokenTypes.EOL),
    ]



from unittest.mock import Mock

from mau.nodes.nodes import Node, ValueNode


def test_node():
    node = Node()

    assert node.node_type == "node"
    assert node._content == {"type": "node"}


def test_value_node():
    node = ValueNode("somevalue")

    assert node.value == "somevalue"
    assert node.node_type == "value_node"
    assert node == ValueNode("somevalue")


def test_node_value_is_none_by_default():
    node = ValueNode()

    assert node.value is None
    assert node == ValueNode()


def test_node_accept():
    visitor = Mock()
    node = Node()
    node.node_type = "mynode"

    assert node.accept(visitor) == visitor._visit_mynode()


def test_node_accept_default():
    # This makes the Mock raise an AttributeError
    # for anything that is not in the list
    visitor = Mock(spec=["_visit_default"])
    node = Node()
    node.node_type = "mynode"

    assert node.accept(visitor) == visitor._visit_default()



from mau.lexers.main_lexer import MainLexer
from mau.nodes.inline import SentenceNode, TextNode
from mau.nodes.page import ContentImageNode, ContentNode
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_include_standard_content():
    source = """
    << somecontent:attr1,attr2,#atag,attr3=value3
    """

    assert runner(source).nodes == [
        ContentNode(
            "somecontent",
            args=["attr1", "attr2"],
            tags=["atag"],
            kwargs={"attr3": "value3"},
        )
    ]


def test_include_image_with_only_path():
    source = """
    << image:/path/to/it.jpg
    """

    assert runner(source).nodes == [ContentImageNode("/path/to/it.jpg")]


def test_include_image_with_http():
    source = """
    << image:https:///some.domain/path/to/it.jpg
    """

    assert runner(source).nodes == [
        ContentImageNode("https:///some.domain/path/to/it.jpg")
    ]


def test_include_image_with_alt_text():
    source = """
    << image:/path/to/it.jpg,alt_text="This is a beautiful image"
    """

    assert runner(source).nodes == [
        ContentImageNode("/path/to/it.jpg", alt_text="This is a beautiful image")
    ]


def test_include_image_with_classes():
    source = """
    << image:/path/to/it.jpg,classes="class1,class2"
    """

    assert runner(source).nodes == [
        ContentImageNode("/path/to/it.jpg", classes=["class1", "class2"])
    ]


def test_include_image_with_arguments():
    source = """
   << image:/path/to/it.jpg,"alternate text", argument1=value1,argument2=value2
   """

    assert runner(source).nodes == [
        ContentImageNode(
            "/path/to/it.jpg",
            args=[],
            kwargs={"argument1": "value1", "argument2": "value2"},
            alt_text="alternate text",
        )
    ]


def test_include_image_with_tags():
    source = """
   << image:/path/to/it.jpg,"alternate text", #tag1, #tag2, argument1=value1,argument2=value2
   """

    assert runner(source).nodes == [
        ContentImageNode(
            "/path/to/it.jpg",
            args=[],
            tags=["tag1", "tag2"],
            kwargs={"argument1": "value1", "argument2": "value2"},
            alt_text="alternate text",
        )
    ]


def test_include_image_with_title():
    source = """
    . A nice caption
    << image:/path/to/it.jpg
    """

    assert runner(source).nodes == [
        ContentImageNode(
            "/path/to/it.jpg",
            title=SentenceNode(
                [
                    TextNode("A nice caption"),
                ]
            ),
        )
    ]



import copy
import logging

import jinja2

# from mau.parsers.toc import create_toc
from mau.visitors.base_visitor import BaseVisitor

_logger = logging.getLogger(__name__)


class TemplateNotFound(ValueError):
    pass


class JinjaVisitor(BaseVisitor):
    format_code = "jinja"
    extension = "j2"
    transform = None

    environment_options = {}
    default_templates = {}

    def __init__(
        self,
        *args,
        custom_templates=None,
        templates_directory=None,
        config=None,
        **kwds,
    ):
        super().__init__()

        self.templates_directory = templates_directory
        self.config = copy.deepcopy(config) if config else {}

        self.templates = {}
        self.templates.update(self.default_templates)

        if custom_templates:
            self.templates.update(custom_templates)

        self._join_with = {
            "block": "\n",
            "command_footnotes": "",
            "command_references": "",
            "command_toc": "",
            "container": "",
            "document": "",
            "footnote": "\n",
            "footnotes_entry": "\n",
            "list": "",
            "reference": "\n",
            "references_entry": "\n",
            "sentence": "",
            "toc_entry": "",
        }

        # These act as a temporary storage while we are in mau blocks.
        # There, ToC and footnotes have to be isolated, so at the
        # beginning of the block we save the main ToC and footnotes
        self._temp_toc = None
        self._temp_footnotes = None

        # This is the environment that uses the internal dict
        self._dict_env = jinja2.Environment(
            loader=jinja2.DictLoader(self.templates),
            **self.environment_options,
        )

        # This is the environment that uses files
        # If the templates directory is not defined we fall back to the dict environment
        if self.templates_directory:  # pragma: no cover
            self._files_env = jinja2.Environment(
                loader=jinja2.FileSystemLoader(searchpath=self.templates_directory),
                **self.environment_options,
            )
        else:
            self._files_env = self._dict_env

    def _render(self, template_full_name, **kwargs):
        # This renders a template using the current config and the given parameters

        try:
            template = self._files_env.get_template(template_full_name)
        except jinja2.exceptions.TemplateNotFound:
            try:
                template = self._dict_env.get_template(template_full_name)
            except jinja2.exceptions.TemplateNotFound as exception:
                raise TemplateNotFound(exception) from exception

        return template.render(config=self.config, **kwargs)

    def visit(self, node, *args, **kwargs):
        # The visitor has to define functions for each node type
        # and those shall return a dictionary of keys.
        #
        # The node types are a list made of the key `node_types` and the node type.
        # This allows a function to return one or more types that have to
        # be used instead of the standard type, which also allows to write
        # generic functions. See _visit_style in HTMLVisitor which works for
        # multiple styles.
        #
        # The rest of the returned keys are passed to the _render function
        # as keys and are thus available in the template.

        if node is None:
            return {}

        result = super().visit(node, *args, **kwargs)

        # The key "templates" contains a list of all templates
        # used for this node.

        # The node type is appended
        # as the last choice to allow specialised
        # templates to be applied first
        templates = result.get("templates", [])
        templates.append(node.node_type)

        # The template full name contains the extension
        # with the type of template, e.g. document.txt
        template_full_names = [f"{template}.{self.extension}" for template in templates]

        # The key "data" contains the values that we want
        # to pass to the template. These are used as arguments
        # when calling the template.
        data = result["data"]

        for template_full_name in template_full_names:
            try:
                return self._render(template_full_name, **data)
            except TemplateNotFound:
                continue

        self._error(
            f"Cannot find any of the following templates {template_full_names}", node
        )

        return None  # pragma: no cover

    def _visit_style(self, node):
        base = super()._visit_style(node)
        base["templates"] = [node.value]

        return base

    #     def _visit_macro(self, node):
    #         return {
    #             "data": {
    #                 "name": node.name,
    #                 "args": node.args,
    #                 "kwargs": node.kwargs,
    #             },
    #             "templates": [f"macro-{node.name}", "macro"],
    #         }

    #     def _visit_header(self, node):
    #         return {
    #             "data": {
    #                 "value": node.value,
    #                 "level": int(node.level),
    #                 "anchor": node.anchor,
    #                 "args": node.args,
    #                 "kwargs": node.kwargs,
    #                 "tags": node.tags,
    #             },
    #             "templates": [f"header{node.level}", "header"],
    #         }

    def _visit_content(self, node):
        base = super()._visit_content(node)
        base["templates"] = [
            f"content-{node.content_type}",
            "content",
        ]

        return base

    def _visit_block(self, node):
        base = super()._visit_block(node)
        base["templates"] = [
            f"block-{node.engine}-{node.blocktype}",
            f"block-{node.engine}",
            f"block-{node.blocktype}",
            "block",
        ]

        return base

    def _visit_command_toc(self, node):
        base = super()._visit_command_toc(node)
        base["templates"] = ["toc"]

        return base

    def _visit_command_footnotes(self, node):
        base = super()._visit_command_footnotes(node)
        base["templates"] = ["footnotes"]

        return base

    def _visit_source(self, node):
        base = super()._visit_source(node)
        base["templates"] = [
            f"source-{node.blocktype}-{node.language}",
            f"source-{node.language}",
            f"source-{node.blocktype}",
            "source",
        ]

        return base

    def _visit_reference(self, node):
        base = super()._visit_reference(node)
        base["templates"] = [
            f"reference-{node.content_type}-{node.category}",
            f"reference-{node.content_type}",
            "reference",
        ]

        return base

    def _visit_command_references(self, node):
        base = super()._visit_command_references(node)
        base["templates"] = [
            f"references-{node.content_type}",
            "references",
        ]

        return base

    def _visit_references_entry(self, node):
        base = super()._visit_references_entry(node)
        base["templates"] = [
            f"references_entry-{node.content_type}-{node.category}",
            f"references_entry-{node.content_type}",
            "references_entry",
        ]

        return base



from mau.nodes.nodes import Node
from mau.nodes.page import PageNode


class CalloutsEntryNode(Node):
    # This is an entry in the list of callouts after source code

    node_type = "callouts_entry"

    def __init__(self, marker, value):
        super().__init__()
        self.marker = marker
        self.value = value

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "marker": self.marker,
            "value": self.value,
        }


class CalloutNode(Node):
    # This is a marker near a source code line

    node_type = "callout"

    def __init__(self, line, marker):
        super().__init__()
        self.line = line
        self.marker = marker

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "line": self.line,
            "marker": self.marker,
        }


class SourceNode(PageNode):
    """A block of verbatim text or source code.

    This node contains verbatim text or source code.

    Arguments:
        language: the language of the code contained in this block
        callouts: a list of callout CalloutEntryNode objects
        markers: list of CalloutNode objects
        highlights: list of lines that have to be highlighted
        delimiter: callouts delimiter
        code: content of the block
        title: title of this block
        kwargs: named arguments
    """

    node_type = "source"

    def __init__(
        self,
        blocktype="default",
        code=None,
        language="text",
        callouts=None,
        delimiter=":",
        markers=None,
        highlights=None,
        classes=None,
        title=None,
        preprocessor=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(args, kwargs, tags)
        self.blocktype = blocktype
        self.code = code or []
        self.language = language
        self.callouts = callouts or []
        self.markers = markers or []
        self.delimiter = delimiter
        self.highlights = highlights or []
        self.classes = classes or []
        self.title = title
        self.preprocessor = preprocessor

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "blocktype": self.blocktype,
            "code": self.code,
            "language": self.language,
            "callouts": self.callouts,
            "markers": self.markers,
            "delimiter": self.delimiter,
            "highlights": self.highlights,
            "classes": self.classes,
            "title": self.title,
            "preprocessor": self.preprocessor,
            "args": self.args,
            "kwargs": self.kwargs,
            "tags": self.tags,
        }



from mau.lexers.main_lexer import MainLexer
from mau.nodes.inline import SentenceNode, StyleNode, TextNode
from mau.nodes.page import ParagraphNode
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_parse_variable_definition_without_value_is_loaded_as_boolean():
    source = ":attr:"
    parser = runner(source)

    assert parser.nodes == []
    assert parser.environment.asdict() == {"attr": True}


def test_parse_variable_definition_without_value_can_be_negated():
    source = ":!attr:"
    parser = runner(source)

    assert parser.nodes == []
    assert parser.environment.asdict() == {"attr": False}


def test_parse_variable_definition_negative_flag_ignores_value():
    source = ":!attr:42"
    parser = runner(source)

    assert parser.nodes == []
    assert parser.environment.asdict() == {"attr": False}


def test_parse_variable_definition_with_value_is_loaded():
    source = ":attr:42"
    parser = runner(source)

    assert parser.nodes == []
    assert parser.environment.asdict() == {"attr": "42"}


def test_parse_variable_definition_value_can_be_any_text():
    source = ":attr:[footnote](http://some.domain/path)"
    parser = runner(source)

    assert parser.nodes == []
    assert parser.environment.asdict() == {
        "attr": "[footnote](http://some.domain/path)",
    }


def test_parse_variable_definition_with_namespace():
    source = ":meta.attr:42"
    parser = runner(source)

    assert parser.nodes == []
    assert parser.environment.asdict() == {"meta": {"attr": "42"}}


def test_parse_variable_definition_with_multiple_dots():
    source = ":meta.category.attr:42"
    parser = runner(source)

    assert parser.nodes == []
    assert parser.environment.asdict() == {"meta": {"category": {"attr": "42"}}}


def test_variable_replacement():
    source = """
    :answer:42

    The answer is {answer}
    """

    parser = runner(source)

    assert parser.nodes == [
        ParagraphNode(
            SentenceNode(
                [TextNode("The answer is 42")],
            )
        )
    ]


def test_variable_replacement_with_namespace():
    source = """
    :mau.answer:42

    The answer is {mau.answer}
    """

    parser = runner(source)

    assert parser.nodes == [
        ParagraphNode(
            SentenceNode(
                [TextNode("The answer is 42")],
            )
        )
    ]


def test_variable_definition_escape():
    source = r"""
    \:answer:42
    """

    parser = runner(source)

    assert parser.nodes == [
        ParagraphNode(
            SentenceNode(
                [TextNode(":answer:42")],
            )
        )
    ]


def test_skip_variable_replacement():
    source = r"""
    :answer:42

    The answer is \{answer\}
    """

    parser = runner(source)

    assert parser.nodes == [
        ParagraphNode(
            SentenceNode(
                [TextNode("The answer is {answer}")],
            )
        )
    ]


def test_variables_are_preprocessed():
    source = r"""
    :important:*IMPORTANT*

    This is {important}
    """

    parser = runner(source)

    assert parser.nodes == [
        ParagraphNode(
            SentenceNode(
                [
                    TextNode("This is "),
                    StyleNode(
                        "star",
                        SentenceNode(
                            [
                                TextNode("IMPORTANT"),
                            ]
                        ),
                    ),
                ],
            )
        )
    ]



from mau.nodes.nodes import Node, ValueNode


class UnnamedArgumentNode(ValueNode):
    """
    This node contains an unnamed argument.
    """

    node_type = "unnamed_argument"


class NamedArgumentNode(Node):
    """
    This node contains a named argument.
    """

    node_type = "named_argument"

    def __init__(self, key, value):
        super().__init__()
        self.key = key
        self.value = value

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "key": self.key,
            "value": self.value,
        }



def set_names_and_defaults(
    args,
    kwargs,
    positional_names,
    default_values=None,
):
    """
    Gives names to positional arguments and assigns
    default values to the ones that have not been
    initialised.
    """

    if default_values is not None:
        _default_values = default_values.copy()
    else:
        _default_values = {}

    # If we pass more positional values than names,
    # some of them won't be converted and become flags
    remaining_args = args[len(positional_names) :]

    positional_arguments = dict(zip(positional_names, args))

    # Named arguments win over the defaults
    _default_values.update(kwargs)

    # Positional arguments with win over all the rest
    _default_values.update(positional_arguments)

    # Positional arguments are mandatory and strict
    # so all the names have to be present in the
    # final dictionary.
    if not set(positional_names).issubset(set(_default_values.keys())):
        raise ValueError(
            f"The following attributes need to be specified: {positional_names}"
        )

    return remaining_args, _default_values



from unittest.mock import patch

from mau.lexers.main_lexer import MainLexer
from mau.nodes.page import BlockNode, CommandTocNode, HeaderNode
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


@patch("mau.parsers.main_parser.header_anchor")
def test_engine_mau(mock_header_anchor):
    mock_header_anchor.return_value = "XXYY"

    source = """
    = Header 1

    = Header 2

    [sometype, engine=mau]
    ----
    = Header 3

    = Header 4
    ----
    """

    par = runner(source)

    assert par.nodes == [
        HeaderNode("Header 1", "1", "XXYY"),
        HeaderNode("Header 2", "1", "XXYY"),
        BlockNode(
            blocktype="sometype",
            content=[
                HeaderNode("Header 3", "1", "XXYY"),
                HeaderNode("Header 4", "1", "XXYY"),
            ],
            secondary_content=[],
            classes=[],
            title=None,
            engine="mau",
            preprocessor="none",
            args=[],
            kwargs={},
        ),
    ]

    assert par.headers == [
        HeaderNode("Header 1", "1", "XXYY"),
        HeaderNode("Header 2", "1", "XXYY"),
    ]


@patch("mau.parsers.main_parser.header_anchor")
def test_engine_mau_multiple_blocks_are_independent(mock_header_anchor):
    mock_header_anchor.return_value = "XXYY"

    source = """
    = Header 1

    = Header 2

    [sometype, engine=mau]
    ----
    = Header 3

    = Header 4
    ----

    [sometype, engine=mau]
    ----
    = Header 5

    = Header 6
    ----
    """

    par = runner(source)

    assert par.nodes == [
        HeaderNode("Header 1", "1", "XXYY"),
        HeaderNode("Header 2", "1", "XXYY"),
        BlockNode(
            blocktype="sometype",
            content=[
                HeaderNode("Header 3", "1", "XXYY"),
                HeaderNode("Header 4", "1", "XXYY"),
            ],
            secondary_content=[],
            classes=[],
            title=None,
            engine="mau",
            preprocessor="none",
            args=[],
            kwargs={},
        ),
        BlockNode(
            blocktype="sometype",
            content=[
                HeaderNode("Header 5", "1", "XXYY"),
                HeaderNode("Header 6", "1", "XXYY"),
            ],
            secondary_content=[],
            classes=[],
            title=None,
            engine="mau",
            preprocessor="none",
            args=[],
            kwargs={},
        ),
    ]

    assert par.headers == [
        HeaderNode("Header 1", "1", "XXYY"),
        HeaderNode("Header 2", "1", "XXYY"),
    ]


@patch("mau.parsers.main_parser.header_anchor")
def test_engine_mau_toc(mock_header_anchor):
    mock_header_anchor.return_value = "XXYY"

    source = """
    = Header 1

    = Header 2

    [sometype, engine=mau]
    ----
    = Header 3

    = Header 4

    ::toc:
    ----

    ::toc:
    """

    par = runner(source)

    assert par.nodes == [
        HeaderNode("Header 1", "1", "XXYY"),
        HeaderNode("Header 2", "1", "XXYY"),
        BlockNode(
            blocktype="sometype",
            content=[
                HeaderNode("Header 3", "1", "XXYY"),
                HeaderNode("Header 4", "1", "XXYY"),
                CommandTocNode(
                    [
                        HeaderNode("Header 3", "1", "XXYY"),
                        HeaderNode("Header 4", "1", "XXYY"),
                    ]
                ),
            ],
            secondary_content=[],
            classes=[],
            title=None,
            engine="mau",
            preprocessor="none",
            args=[],
            kwargs={},
        ),
        CommandTocNode(
            [
                HeaderNode("Header 1", "1", "XXYY"),
                HeaderNode("Header 2", "1", "XXYY"),
            ]
        ),
    ]

    assert par.headers == [
        HeaderNode("Header 1", "1", "XXYY"),
        HeaderNode("Header 2", "1", "XXYY"),
    ]



from mau.lexers.base_lexer import BaseLexer, TokenTypes


class PreprocessVariablesLexer(BaseLexer):
    def _process_literal(self):
        if self._current_char not in r"\`{}":
            return None

        return [self._create_token_and_skip(TokenTypes.LITERAL, self._current_char)]

    def _process_text(self):
        return [self._create_token_and_skip(TokenTypes.TEXT, self._current_char)]

    def _process_functions(self):
        return [
            self._process_literal,
            self._process_text,
        ]



from mau.text_buffer.text_buffer import Context, TextBuffer

from tests.helpers import dedent


def test_text_buffer_current_line():
    text_buffer = TextBuffer("abcdef")

    assert text_buffer.current_line == "abcdef"


def test_text_buffer_current_line_goes_beyond_eof():
    text_buffer = TextBuffer("abcdef")
    text_buffer.line = 10

    assert text_buffer.current_line == ""


def test_text_buffer_current_char():
    text_buffer = TextBuffer("abcdef")

    assert text_buffer.current_char == "a"


def test_text_buffer_current_char_goes_beyond_eol():
    text_buffer = TextBuffer("abcdef")
    text_buffer.column = 10

    assert text_buffer.current_char == ""


def test_text_buffer_current_char_goes_beyond_eof():
    text_buffer = TextBuffer("abcdef")
    text_buffer.line = 10

    assert text_buffer.current_char == ""


def test_text_buffer_eof():
    text_buffer = TextBuffer("abcdef")

    text_buffer.line = 0
    assert text_buffer.eof is False

    text_buffer.line = 1
    assert text_buffer.eof is True

    text_buffer.line = 10
    assert text_buffer.eof is True


def test_text_buffer_eol():
    text_buffer = TextBuffer("abcdef", Context(source="main"))

    text_buffer.column = 0
    assert text_buffer.eol is False

    text_buffer.column = 6
    assert text_buffer.eol is True

    text_buffer.column = 10
    assert text_buffer.eol is True


def test_text_buffer_eof_eol_empty_file():
    text_buffer = TextBuffer("")

    assert text_buffer.eof is True
    assert text_buffer.eol is True


def test_text_buffer_peek_char():
    text_buffer = TextBuffer("abcdef")

    assert text_buffer.peek_char == "b"


def test_text_buffer_peek_char_goes_beyond_eol():
    text_buffer = TextBuffer("abcdef")
    text_buffer.column = 10

    assert text_buffer.peek_char == ""


def test_text_buffer_peek_char_goes_beyond_eof():
    text_buffer = TextBuffer("abcdef")
    text_buffer.line = 1

    assert text_buffer.peek_char == ""


def test_text_buffer_tail():
    text_buffer = TextBuffer("abcdefgh")
    text_buffer.column = 4

    assert text_buffer.tail == "efgh"


def test_text_buffer_tail_works_after_eol():
    text_buffer = TextBuffer("abcdefgh")
    text_buffer.column = 100

    assert text_buffer.tail == ""


def test_text_buffer_tail_works_after_eof():
    text_buffer = TextBuffer("abcdefgh")
    text_buffer.line = 10

    assert text_buffer.tail == ""


def test_text_buffer_nextline():
    text_buffer = TextBuffer("abc\ndef\nghi")
    text_buffer.nextline()

    assert text_buffer.line == 1
    assert text_buffer.column == 0


def test_text_buffer_nextline_last_line():
    text_buffer = TextBuffer("abc\ndef\nghi")
    text_buffer.line = 3
    text_buffer.column = 2
    text_buffer.nextline()

    assert text_buffer.line == 4
    assert text_buffer.column == 0


def test_text_buffer_position():
    text_buffer = TextBuffer()
    text_buffer.line = 1
    text_buffer.column = 3

    assert text_buffer.position == (1, 3)


def test_text_buffer_position_outext_bufferide():
    text_buffer = TextBuffer()
    text_buffer.line = 123
    text_buffer.column = 456

    assert text_buffer.position == (123, 456)


def test_text_buffer_set_position():
    text_buffer = TextBuffer()
    text_buffer.position = (222, 333)

    assert text_buffer.position == (222, 333)


def test_text_buffer_skip_defaults_to_one():
    text_buffer = TextBuffer("abc\ndef\nghi")
    text_buffer.skip()

    assert text_buffer.column == 1


def test_text_buffer_skip_accepts_value():
    text_buffer = TextBuffer("abc\ndef\nghi")
    text_buffer.skip(3)

    assert text_buffer.column == 3


def test_text_buffer_skip_can_go_beyond_eol():
    text_buffer = TextBuffer("abc\ndefghi")
    text_buffer.nextline()
    text_buffer.skip(25)

    assert text_buffer.column == 6


def test_text_buffer_default_context():
    text_buffer = TextBuffer(
        dedent(
            """
            Some text
            on multiple lines
            that I will use
            to check the context()
            """
        ),
    )

    text_buffer.position = (2, 6)

    assert text_buffer.context.asdict() == {
        "source": None,
        "text": "that I will use",
        "line": 2,
        "column": 6,
    }


def test_text_buffer_initial_context():
    text_buffer = TextBuffer(
        dedent(
            """
            Some text
            on multiple lines
            that I will use
            to check the context()
            """
        ),
        Context(72, 123, "main"),
    )

    assert text_buffer.current_line == "Some text"
    assert text_buffer.line == 0
    assert text_buffer.column == 0
    assert text_buffer.context.asdict() == {
        "source": "main",
        "text": "Some text",
        "line": 72,
        "column": 123,
    }



# Page nodes can be found at top level in a page

from mau.nodes.nodes import Node


class PageNode(Node):
    node_type = "page_node"

    def __init__(
        self,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__()
        self.args = args or []
        self.kwargs = kwargs or {}
        self.tags = tags or []

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "args": self.args,
            "kwargs": self.kwargs,
            "tags": self.tags,
        }


class HorizontalRuleNode(PageNode):
    """A horizontal rule."""

    node_type = "horizontal_rule"


class ParagraphNode(PageNode):
    """A paragraph."""

    node_type = "paragraph"

    def __init__(
        self,
        content,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(args, kwargs, tags)
        self.content = content

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "content": self.content,
            "args": self.args,
            "kwargs": self.kwargs,
            "tags": self.tags,
        }


class CommandTocNode(PageNode):
    """A Table of Contents command.

    This node contains the headers that go into the ToC.
    """

    node_type = "command_toc"

    def __init__(self, entries, args=None, kwargs=None, tags=None):
        super().__init__(args, kwargs, tags)
        self.entries = entries

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "entries": self.entries,
            "args": self.args,
            "kwargs": self.kwargs,
            "tags": self.tags,
        }


class HeaderNode(PageNode):
    """A header."""

    node_type = "header"

    def __init__(
        self,
        value,
        level,
        anchor,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(args, kwargs, tags)
        self.value = value
        self.level = level
        self.anchor = anchor

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "value": self.value,
            "level": self.level,
            "anchor": self.anchor,
            "args": self.args,
            "kwargs": self.kwargs,
            "tags": self.tags,
        }


class ListNode(PageNode):
    """A list."""

    node_type = "list"

    def __init__(
        self,
        ordered,
        items,
        main_node=False,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(args, kwargs, tags)
        self.ordered = ordered
        self.items = items
        self.main_node = main_node

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "ordered": self.ordered,
            "items": self.items,
            "main_node": self.main_node,
            "args": self.args,
            "kwargs": self.kwargs,
            "tags": self.tags,
        }


class ContentNode(PageNode):
    """Content included in the page.

    This represents generic content included in the page.

    Arguments:
        content_type: the type of content
        title: caption of the image
        args: unnamed arguments
        kwargs: named arguments
    """

    node_type = "content"

    def __init__(
        self,
        content_type,
        title=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(args, kwargs, tags)
        self.content_type = content_type
        self.title = title

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "content_type": self.content_type,
            "title": self.title,
            "args": self.args,
            "kwargs": self.kwargs,
            "tags": self.tags,
        }


class ContentImageNode(PageNode):
    """An image included in the page."""

    node_type = "content_image"

    def __init__(
        self,
        uri,
        alt_text=None,
        classes=None,
        title=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(args, kwargs, tags)
        self.uri = uri
        self.classes = classes
        self.title = title
        self.alt_text = alt_text

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "uri": self.uri,
            "alt_text": self.alt_text,
            "classes": self.classes,
            "title": self.title,
            "args": self.args,
            "kwargs": self.kwargs,
            "tags": self.tags,
        }


class BlockNode(PageNode):
    """A block.

    This node contains a generic block.

    Arguments:
        blocktype: the type of this block
        content: content of the block
        secondary_content: secondary content of this block
        title: title of this block
        classes: a comma-separated list of classes
        engine: the engine used to render this block
        preprocessor: the preprocessor used for this block
        args: unnamed arguments
        kwargs: named arguments
    """

    node_type = "block"

    def __init__(
        self,
        blocktype,
        content,
        secondary_content,
        classes=None,
        title=None,
        engine=None,
        preprocessor=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(args, kwargs, tags)
        self.blocktype = blocktype
        self.content = content
        self.secondary_content = secondary_content
        self.title = title
        self.classes = classes or []
        self.engine = engine
        self.preprocessor = preprocessor

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "blocktype": self.blocktype,
            "content": self.content,
            "secondary_content": self.secondary_content,
            "title": self.title,
            "classes": self.classes,
            "engine": self.engine,
            "preprocessor": self.preprocessor,
            "args": self.args,
            "kwargs": self.kwargs,
            "tags": self.tags,
        }


class ContainerNode(PageNode):
    node_type = "container"

    def __init__(
        self,
        content,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(args, kwargs, tags)
        self.content = content

    @property
    def _content(self):
        return {
            "type": self.node_type,
            "content": self.content,
            "args": self.args,
            "kwargs": self.kwargs,
            "tags": self.tags,
        }


class DocumentNode(ContainerNode):
    """A document.

    This node represents the full document.

    Arguments:
        content: the content of the document
    """

    node_type = "document"



from mau.nodes.toc import TocEntryNode, TocNode


def test_toc_entry_node():
    node = TocEntryNode(value="Header 1", anchor="header-1", children=[])

    assert node.value == "Header 1"
    assert node.anchor == "header-1"
    assert node.children == []
    assert node.node_type == "toc_entry"
    assert node == TocEntryNode(value="Header 1", anchor="header-1", children=[])


def test_toc_node():
    node = TocNode(
        entries=[],
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )

    assert node.entries == []
    assert node.args == ["value1", "value2"]
    assert node.tags == ["tag1", "tag2"]
    assert node.kwargs == {"key1": "text1", "key2": "text2"}
    assert node.node_type == "toc"
    assert node == TocNode(
        entries=[],
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )



from mau.nodes.toc import TocEntryNode, TocNode


def create_toc(headers, *args, **kwargs):
    # Create the TOC from the list of headers.

    nodes = []
    latest_by_level = {}

    for header_node in headers:
        # This is the current node
        node = TocEntryNode(
            header_node.value,
            header_node.anchor,
            args=header_node.args,
            kwargs=header_node.kwargs,
            tags=header_node.tags,
        )
        level = int(header_node.level)

        # This collects the latest node added with a given level
        latest_by_level[level] = node

        try:
            # Simplest case, add it to the latest one
            # with a level just 1 step lower
            latest_by_level[level - 1].children.append(node)
        except KeyError:
            # Find all the latest ones added with a level lower than this
            latest = [latest_by_level.get(i, None) for i in range(1, level)]

            # Get the children list of each one, plus nodes for the root
            children = [nodes] + [i.children for i in latest if i is not None]

            # Get the nearest one and append to that
            children[-1].append(node)

    return TocNode(entries=nodes, *args, **kwargs)



import textwrap
from unittest.mock import patch

from mau.lexers.main_lexer import MainLexer
from mau.nodes.footnotes import FootnoteNode
from mau.nodes.inline import SentenceNode, TextNode
from mau.nodes.page import ParagraphNode
from mau.parsers.main_parser import MainParser, footnote_anchor, reference_anchor

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


@patch("mau.parsers.main_parser.hashlib.md5")
def test_default_footnote_anchor_function(mock_md5):
    mock_md5().hexdigest.return_value = "XXYYXXYYZZZ"

    assert footnote_anchor("Some Words 1234 56") == "XXYYXXYY"


@patch("mau.parsers.main_parser.hashlib.md5")
def test_default_reference_anchor_function(mock_md5):
    mock_md5().hexdigest.return_value = "XXYYXXYYZZZ"

    assert reference_anchor("Some Words 1234 56") == "XXYYXXYY"


def test_footnote_empty_block():
    source = """
    [footnote, note1]
    ----
    ----
    """

    parser = init_parser(textwrap.dedent(source))
    parser.footnotes = {
        "note1": FootnoteNode(number=1),
        "note2": FootnoteNode(number=2),
    }
    parser.parse()

    assert parser.nodes == []


@patch("mau.parsers.main_parser.footnote_anchor")
def test_footnote_content(mock_footnote_anchor):
    mock_footnote_anchor.return_value = "XXYY"

    source = """
    [footnote, note2]
    ----
    This is another paragraph.
    ----

    [footnote, note1]
    ----
    This is a paragraph.
    ----
    """

    parser = init_parser(textwrap.dedent(source))
    parser.footnotes = {
        "note1": FootnoteNode(number=1),
        "note2": FootnoteNode(number=2),
    }
    parser.parse()
    parser.create_footnotes()

    assert parser.nodes == []
    assert parser.footnotes == {
        "note1": FootnoteNode(
            number=1,
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("This is a paragraph."),
                        ]
                    )
                ),
            ],
            reference_anchor="fr-XXYY-1",
            content_anchor="fd-XXYY-1",
        ),
        "note2": FootnoteNode(
            number=2,
            content=[
                ParagraphNode(
                    SentenceNode(
                        [
                            TextNode("This is another paragraph."),
                        ]
                    )
                ),
            ],
            reference_anchor="fr-XXYY-2",
            content_anchor="fd-XXYY-2",
        ),
    }



from mau.tokens.tokens import Context, Token


def test_token_accepts_type_and_value():
    token = Token("sometype", "somevalue")

    assert token.type == "sometype"
    assert token.value == "somevalue"
    assert token != "somevalue"


def test_token_keeps_value_none():
    token = Token("sometype", None)

    assert token.type == "sometype"
    assert token.value == ""


def test_token_value_defaults_to_none():
    token = Token("sometype")

    assert token.type == "sometype"
    assert token.value == ""


def test_token_equality():
    assert Token("sometype", "somevalue") == Token("sometype", "somevalue")


def test_token_length():
    token = Token("sometype", "somevalue")

    assert len(token) == len("somevalue")
    assert bool(token) is True


def test_empty_token_has_length_zero():
    token = Token("sometype")

    assert len(token) == 0
    assert bool(token) is True


def test_token_accepts_context():
    context = Context(source="main", text="sometext", line=123, column=456)

    token = Token("sometype", "somevalue", context=context)

    assert token.type == "sometype"
    assert token.value == "somevalue"
    assert token.context == context


def test_token_equality_ignores_context():
    context = Context(source="main", text="sometext", line=123, column=456)

    assert Token("sometype", "somevalue", context=context) == Token(
        "sometype", "somevalue"
    )
    assert Token("sometype", "somevalue") == Token(
        "sometype", "somevalue", context=context
    )


def test_token_equality_accepts_none():
    assert Token("sometype", "somevalue") is not None


def test_token_equality_with_any():
    assert Token("sometype", "somevalue") == Token("sometype")


def test_token_match_considers_context():
    context = Context(source="main", text="sometext", line=123, column=456)

    assert not Token("sometype", "somevalue", context=context).match(
        Token("sometype", "somevalue")
    )
    assert not Token("sometype", "somevalue").match(
        Token("sometype", "somevalue", context=context)
    )
    assert Token("sometype", "somevalue", context=context).match(
        Token("sometype", "somevalue", context=context)
    )



import pytest
from mau.lexers.main_lexer import MainLexer
from mau.nodes.inline import SentenceNode, TextNode
from mau.nodes.source import CalloutNode, CalloutsEntryNode, SourceNode
from mau.parsers.base_parser import ParserError
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

# from mau.parsers.base_parser import ParserError


init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_source_empty_block():
    source = """
    [source]
    ----
    ----
    """

    assert runner(source).nodes == [
        SourceNode(blocktype="default", code=[]),
    ]


def test_source_language():
    source = """
    [source, python]
    ----
    ----
    """

    assert runner(source).nodes == [
        SourceNode(code=[], language="python"),
    ]


def test_source_contains_mau():
    source = """
    [source, callouts="|"]
    ----
    // A comment
    ::toc:
    ----
    """

    assert runner(source).nodes == [
        SourceNode(
            code=[
                TextNode("// A comment"),
                TextNode("::toc:"),
            ],
        ),
    ]


def test_source_removes_escape_from_directive_like_text():
    source = r"""
    [source]
    ----
    \::#looks like a directive
    ----
    """

    assert runner(source).nodes == [
        SourceNode(
            code=[
                TextNode("::#looks like a directive"),
            ],
        ),
    ]


def test_source_with_code():
    source = """
    [source]
    ----
    import os

    print(os.environ["HOME"])
    ----
    """

    assert runner(source).nodes == [
        SourceNode(
            code=[
                TextNode("import os"),
                TextNode(""),
                TextNode('print(os.environ["HOME"])'),
            ],
        ),
    ]


def test_source_explicit_engine():
    source = """
    [myblock, engine=source, language=somelang]
    ----
    import os

    print(os.environ["HOME"])
    ----
    """

    assert runner(source).nodes == [
        SourceNode(
            blocktype="myblock",
            language="somelang",
            code=[
                TextNode("import os"),
                TextNode(""),
                TextNode('print(os.environ["HOME"])'),
            ],
        ),
    ]


def test_source_with_title():
    source = """
    . Title
    [source, somelang]
    ----
    import os

    print(os.environ["HOME"])
    ----
    """

    assert runner(source).nodes == [
        SourceNode(
            language="somelang",
            code=[
                TextNode("import os"),
                TextNode(""),
                TextNode('print(os.environ["HOME"])'),
            ],
            title=SentenceNode([TextNode("Title")]),
        ),
    ]


def test_source_ignores_mau_syntax():
    source = """
    [source]
    ----
    :answer:42
    The answer is {answer}
    ----
    """

    assert runner(source).nodes == [
        SourceNode(
            code=[
                TextNode(":answer:42"),
                TextNode("The answer is {answer}"),
            ],
        ),
    ]


def test_source_respects_spaces_and_indentation():
    source = """
    [source]
    ----
      //    This is a comment
    ----
    """

    assert runner(source).nodes == [
        SourceNode(
            code=[
                TextNode("  //    This is a comment"),
            ],
        ),
    ]


def test_source_callouts():
    source = """
    [source, somelang, callouts=":"]
    ----
    import sys
    import os:imp:

    print(os.environ["HOME"]):env:
    ----
    imp: This is an import
    env: Environment variables are paramount
    """

    assert runner(source).nodes == [
        SourceNode(
            language="somelang",
            code=[
                TextNode("import sys"),
                TextNode("import os"),
                TextNode(""),
                TextNode('print(os.environ["HOME"])'),
            ],
            markers=[CalloutNode(1, "imp"), CalloutNode(3, "env")],
            callouts=[
                CalloutsEntryNode("imp", "This is an import"),
                CalloutsEntryNode("env", "Environment variables are paramount"),
            ],
        ),
    ]


def test_source_callouts_possible_clash():
    source = """
    [source, somelang, callouts=":"]
    ----
    import sys
    import: os:imp:

    print(os.environ["HOME"]):env:
    ----
    imp: This is an import
    env: Environment variables are paramount
    """

    assert runner(source).nodes == [
        SourceNode(
            language="somelang",
            code=[
                TextNode("import sys"),
                TextNode("import: os"),
                TextNode(""),
                TextNode('print(os.environ["HOME"])'),
            ],
            markers=[CalloutNode(1, "imp"), CalloutNode(3, "env")],
            callouts=[
                CalloutsEntryNode("imp", "This is an import"),
                CalloutsEntryNode("env", "Environment variables are paramount"),
            ],
        ),
    ]


def test_source_callouts_one_single_marker_is_skipped():
    source = """
    [source, somelang, callouts=":"]
    ----
    def something:
        print("AAA")
    ----
    """

    assert runner(source).nodes == [
        SourceNode(
            language="somelang",
            code=[
                TextNode("def something:"),
                TextNode('    print("AAA")'),
            ],
            markers=[],
            callouts=[],
        ),
    ]


def test_source_callouts_custom_delimiter():
    source = """
    [source, language=somelang, callouts="|"]
    ----
    import sys
    import os|imp|

    print(os.environ["HOME"])|env|
    ----
    """

    assert runner(source).nodes == [
        SourceNode(
            language="somelang",
            code=[
                TextNode("import sys"),
                TextNode("import os"),
                TextNode(""),
                TextNode('print(os.environ["HOME"])'),
            ],
            markers=[CalloutNode(1, "imp"), CalloutNode(3, "env")],
        ),
    ]


def test_source_callout_wrong_format():
    # This is testing that all callouts in the secondary
    # content have the right format "label: content"

    source = """
    [myblock, engine=source, language=somelang, callouts=":"]
    ----
    import sys
    import os:3:

    print(os.environ["HOME"]):6:
    ----
    3 This is an import
    6: Environment variables are paramount
    """

    with pytest.raises(ParserError):
        runner(source)


def test_source_highlights():
    source = """
    [source, language=somelang]
    ----
    import sys
    import os:@:

    print(os.environ["HOME"]):@:
    ----
    """

    assert runner(source).nodes == [
        SourceNode(
            language="somelang",
            code=[
                TextNode("import sys"),
                TextNode("import os"),
                TextNode(""),
                TextNode('print(os.environ["HOME"])'),
            ],
            highlights=[1, 3],
        ),
    ]


def test_source_highlights_custom_marker():
    source = """
    [source, language=somelang, highlight="#"]
    ----
    import sys
    import os:#:

    print(os.environ["HOME"]):#:
    ----
    """

    assert runner(source).nodes == [
        SourceNode(
            language="somelang",
            code=[
                TextNode("import sys"),
                TextNode("import os"),
                TextNode(""),
                TextNode('print(os.environ["HOME"])'),
            ],
            highlights=[1, 3],
        ),
    ]


def test_engine_source_language_is_reset():
    source = """
    [source, python]
    ----
    ----

    [source]
    ----
    ----
    """

    assert runner(source).nodes == [
        SourceNode(
            language="python",
            code=[],
        ),
        SourceNode(
            language="text",
            code=[],
        ),
    ]



import pytest

from mau.nodes.inline import (
    TextNode,
    VerbatimNode,
    SentenceNode,
    StyleNode,
    MacroNode,
    ClassNode,
    LinkNode,
    ImageNode,
    ListItemNode,
)

from mau.visitors.base_visitor import VisitorError
from mau.visitors.jinja_visitor import JinjaVisitor
from mau.nodes.footnotes import FootnoteNode
from mau.nodes.references import ReferenceNode


def test_default_values():
    visitor = JinjaVisitor()

    assert visitor.default_templates == {}
    assert visitor.environment_options == {}
    assert visitor.templates_directory is None
    assert visitor.config == {}
    assert visitor.extension == "j2"
    assert visitor.toc is None
    assert visitor.footnotes is None

    assert visitor.templates == {}


def test_custom_templates():
    sometemplates = {"key": "value"}
    visitor = JinjaVisitor(custom_templates=sometemplates)

    assert visitor.templates == sometemplates


def test_custom_templates_with_default_templates():
    class MyJinjaVisitor(JinjaVisitor):
        default_templates = {"defaultkey": "defaultvalue"}

    some_templates = {"key": "value"}
    merged_templates = {"defaultkey": "defaultvalue", "key": "value"}

    visitor = MyJinjaVisitor(custom_templates=some_templates)

    assert visitor.templates == merged_templates


def test_no_templates():
    templates = {}

    visitor = JinjaVisitor(custom_templates=templates)

    node = TextNode("Just some text.")

    with pytest.raises(VisitorError):
        visitor.visit(node)


def test_inline_text_node():
    templates = {
        "text.j2": "{{ value }}",
    }

    visitor = JinjaVisitor(custom_templates=templates)

    node = TextNode("Just some text.")

    result = visitor.visit(node)

    assert result == "Just some text."


def test_inline_verbatim_node():
    templates = {
        "verbatim.j2": "{{ value }}",
    }

    visitor = JinjaVisitor(custom_templates=templates)

    node = VerbatimNode("Just some text.")

    result = visitor.visit(node)

    assert result == "Just some text."


def test_inline_sentence_node():
    templates = {
        "text.j2": "{{ value }}",
        "sentence.j2": "{{ content }}",
    }

    visitor = JinjaVisitor(custom_templates=templates)

    node = SentenceNode(
        [
            TextNode("Just some text. "),
            TextNode("More text."),
        ]
    )

    result = visitor.visit(node)

    assert result == "Just some text. More text."


def test_inline_style_node_star():
    templates = {
        "text.j2": "{{ value }}",
        "sentence.j2": "{{ content }}",
        "star.j2": "*{{ content }}*",
    }

    visitor = JinjaVisitor(custom_templates=templates)

    node = StyleNode("star", SentenceNode([TextNode("Just some text.")]))

    result = visitor.visit(node)

    assert result == "*Just some text.*"


def test_inline_style_node_underscore():
    templates = {
        "text.j2": "{{ value }}",
        "sentence.j2": "{{ content }}",
        "underscore.j2": "_{{ content }}_",
    }

    visitor = JinjaVisitor(custom_templates=templates)

    node = StyleNode("underscore", SentenceNode([TextNode("Just some text.")]))

    result = visitor.visit(node)

    assert result == "_Just some text._"


def test_inline_style_node_tilde():
    templates = {
        "text.j2": "{{ value }}",
        "sentence.j2": "{{ content }}",
        "tilde.j2": "~{{ content }}~",
    }

    visitor = JinjaVisitor(custom_templates=templates)

    node = StyleNode("tilde", SentenceNode([TextNode("Just some text.")]))

    result = visitor.visit(node)

    assert result == "~Just some text.~"


def test_inline_style_node_caret():
    templates = {
        "text.j2": "{{ value }}",
        "sentence.j2": "{{ content }}",
        "caret.j2": "^{{ content }}^",
    }

    visitor = JinjaVisitor(custom_templates=templates)

    node = StyleNode("caret", SentenceNode([TextNode("Just some text.")]))

    result = visitor.visit(node)

    assert result == "^Just some text.^"


def test_inline_macro_node():
    templates = {
        "text.j2": "{{ value }}",
        "macro.j2": (
            "{{ name }} - {{ args | join(',') }} - "
            "{% for key, value in kwargs|items %}{{ key }}:{{ value }}{% endfor %}"
        ),
    }

    visitor = JinjaVisitor(custom_templates=templates)

    node = MacroNode("somename", ["arg1", "arg2"], {"key1": "value1"})

    result = visitor.visit(node)

    assert result == "somename - arg1,arg2 - key1:value1"


def test_inline_footnote_node():
    templates = {
        "text.j2": "{{ value }}",
        "footnote.j2": (
            "{{ content }} - {{ number }} - "
            "{{ reference_anchor }} - {{ content_anchor }}"
        ),
    }

    visitor = JinjaVisitor(custom_templates=templates)

    node = FootnoteNode(
        [TextNode("Just some text.")], "5", "someanchor", "someanchor-def"
    )

    result = visitor.visit(node)

    assert result == "Just some text. - 5 - someanchor - someanchor-def"


def test_inline_reference_node():
    templates = {
        "text.j2": "{{ value }}",
        "sentence.j2": "{{ content }}",
        "reference-somecontent.j2": (
            "{{ content_type }} - {{ name }} - {{category}} - {{ content }} - "
            "{{ number }} - {{ title }} - {{ reference_anchor }} - {{ content_anchor }}"
        ),
    }

    visitor = JinjaVisitor(custom_templates=templates)

    node = ReferenceNode(
        "somecontent",
        "somevalue",
        "somecategory",
        [TextNode("Just some text.")],
        "5",
        SentenceNode([TextNode("Some title")]),
        "someanchor",
        "someanchor-def",
    )

    result = visitor.visit(node)

    assert result == (
        "somecontent - somevalue - somecategory - Just some text. "
        "- 5 - Some title - someanchor - someanchor-def"
    )


def test_inline_class_node():
    templates = {
        "text.j2": "{{ value }}",
        "class.j2": "{{ classes | join(',') }} - {{ content }}",
    }

    visitor = JinjaVisitor(custom_templates=templates)

    node = ClassNode(["class1", "class2"], TextNode("Just some text."))

    result = visitor.visit(node)

    assert result == "class1,class2 - Just some text."


def test_inline_link_node():
    templates = {
        "text.j2": "{{ value }}",
        "link.j2": "{{ target }} - {{ text }}",
    }

    visitor = JinjaVisitor(custom_templates=templates)

    node = LinkNode(target="sometarget", text="sometext")

    result = visitor.visit(node)

    assert result == "sometarget - sometext"


def test_inline_image_node():
    templates = {
        "text.j2": "{{ value }}",
        "image.j2": "{{ uri }} - {{ alt_text }} - {{ width }}x{{ height }}",
    }

    visitor = JinjaVisitor(custom_templates=templates)

    node = ImageNode(uri="someuri", alt_text="sometext", width="100", height="400")

    result = visitor.visit(node)

    assert result == "someuri - sometext - 100x400"


def test_inline_list_item_node():
    templates = {
        "text.j2": "{{ value }}",
        "list_item.j2": "{{ level }} - {{ content }}",
    }

    visitor = JinjaVisitor(custom_templates=templates)

    node = ListItemNode("4", TextNode("Just some text."))

    result = visitor.visit(node)

    assert result == "4 - Just some text."



from mau.nodes.inline import TextNode
from mau.nodes.source import CalloutNode, CalloutsEntryNode, SourceNode


def test_callouts_entry_node():
    node = CalloutsEntryNode("somemarker", "somevalue")

    assert node.marker == "somemarker"
    assert node.value == "somevalue"
    assert node.node_type == "callouts_entry"
    assert node == CalloutsEntryNode("somemarker", "somevalue")


def test_callout_node():
    node = CalloutNode("2", "somemarker")

    assert node.line == "2"
    assert node.marker == "somemarker"
    assert node.node_type == "callout"
    assert node == CalloutNode("2", "somemarker")


def test_source_node():
    node = SourceNode(
        code=[TextNode("somecode")],
        callouts=[CalloutsEntryNode("somemarker", "somevalue")],
        markers=[CalloutNode("2", "somemarker")],
    )

    assert node.code == [TextNode("somecode")]
    assert node.language == "text"
    assert node.callouts == [CalloutsEntryNode("somemarker", "somevalue")]
    assert node.delimiter == ":"
    assert node.markers == [CalloutNode("2", "somemarker")]
    assert node.highlights == []
    assert node.classes == []
    assert node.title is None
    assert node.preprocessor is None
    assert node.args == []
    assert node.kwargs == {}
    assert node.tags == []
    assert node.node_type == "source"
    assert node == SourceNode(
        code=[TextNode("somecode")],
        callouts=[CalloutsEntryNode("somemarker", "somevalue")],
        markers=[CalloutNode("2", "somemarker")],
    )



from dataclasses import dataclass
from typing import Optional


@dataclass
class Context:
    # Context objects represent the place where a token was found
    # in the source code. They contain line and column where they
    # begin, the name of the source file (if provided), and the
    # text of the full line.

    line: int = 0
    column: int = 0
    source: Optional[str] = None
    text: str = ""

    def asdict(self):
        return {
            "line": self.line,
            "column": self.column,
            "source": self.source,
            "text": self.text,
        }

    def __repr__(self):
        return f"Context({self.line},{self.column},{self.source})"


def print_context(context, error_message=None):  # pragma: no cover
    # This is a function used to print out the context of a token
    # when an error occurs in Mau. It's a typical graphical
    # error-reporting function: a mess.

    len_bar = len(error_message) if error_message is not None else 0
    title = " Mau Error "

    # This takes into account the title and the spaces
    half_bar = "=" * ((len_bar - len(title)) // 2 + 1)
    full_bar = half_bar + "=" * len(title) + half_bar

    print(f"{half_bar}{title}{half_bar}")

    if error_message is not None:
        print(error_message)
        print()

    if context.source is not None:
        print(f"Source: {context.source}")
    print(f"Line: {context.line} - Column: {context.column}")
    print()

    print(f"{context.line}: {context.text}")
    print(" " * (len(str(context.line)) + 2) + "^")
    print(f"{full_bar}")



import pytest
from mau.lexers.preprocess_variables_lexer import PreprocessVariablesLexer
from mau.nodes.inline import TextNode
from mau.parsers.base_parser import ParserError
from mau.parsers.environment import Environment
from mau.parsers.preprocess_variables_parser import PreprocessVariablesParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(PreprocessVariablesLexer, PreprocessVariablesParser)
runner = parser_runner_factory(PreprocessVariablesLexer, PreprocessVariablesParser)


def test_plain_text_with_no_variables():
    source = "This is text"

    assert runner(source).nodes == [TextNode("This is text")]


def test_plain_text_with_variables():
    source = "This is text"
    result = runner(source, environment=Environment({"attr": "5"}))

    assert result.nodes == [TextNode("This is text")]


def test_replace_variable():
    source = "This is number {attr}"
    result = runner(source, environment=Environment({"attr": "5"}))

    assert result.nodes == [TextNode("This is number 5")]


def test_manage_unclosed_curly_braces():
    source = "This is {attr"
    result = runner(source, environment=Environment({"attr": "5"}))

    assert result.nodes == [TextNode("This is {attr")]


def test_replace_variable_with_namespace():
    source = "This is number {config.attr}"
    result = runner(
        source,
        environment=Environment({"config": {"attr": "5"}}),
    )

    assert result.nodes == [TextNode("This is number 5")]


def test_replace_boolean():
    result = runner("This flag is {flag}", environment=Environment({"flag": True}))

    assert result.nodes == [TextNode("This flag is True")]


def test_escape_curly_braces():
    source = r"This is \{attr\}"
    result = runner(source, environment=Environment({"attr": "5"}))

    assert result.nodes == [TextNode("This is {attr}")]


def test_curly_braces_in_verbatim():
    source = "This is `{attr}`"
    result = runner(source, environment=Environment({"attr": "5"}))

    assert result.nodes == [TextNode("This is `{attr}`")]


def test_open_verbatim():
    source = "This is `{attr}"
    result = runner(source, environment=Environment({"attr": "5"}))

    assert result.nodes == [TextNode("This is `5")]


def test_escape_curly_braces_in_verbatim():
    source = r"This is `\{attr\}`"
    result = runner(source, environment=Environment({"attr": "5"}))

    assert result.nodes == [TextNode(r"This is `\{attr\}`")]


def test_escape_other_chars():
    source = r"This \_is\_ \text"
    result = runner(source, environment=Environment({"attr": "5"}))

    assert result.nodes == [TextNode(r"This \_is\_ \text")]


def test_curly_braces_in_escaped_verbatim():
    source = r"This is \`{attr}\`"
    result = runner(source, environment=Environment({"attr": "5"}))

    assert result.nodes == [TextNode(r"This is \`5\`")]


def test_variable_not_existing():
    source = "This is number {attr}"

    with pytest.raises(ParserError):
        runner(source, environment={})


def test_variables_can_contain_markers():
    source = "A very {bold} text. Some code: {dictdef}"
    result = runner(
        source,
        environment=Environment({"bold": "*bold*", "dictdef": "`adict = {'a':5}`"}),
    )

    assert result.nodes == [
        TextNode("A very *bold* text. Some code: `adict = {'a':5}`")
    ]


def test_escape_backtick():
    result = runner(r"This is `\``")

    assert result.nodes == [TextNode(r"This is `\``")]



from mau.nodes.page import HeaderNode
from mau.nodes.toc import TocEntryNode, TocNode
from mau.parsers.toc import create_toc


def test_create_toc():
    headers_list = [
        HeaderNode("Header 1", "1", "header-1"),
        HeaderNode("Header 1.1", "2", "header-1-1"),
        HeaderNode("Header 1.2", "2", "header-1-2"),
        HeaderNode("Header 2", "1", "header-2"),
        HeaderNode("Header 2.1", "2", "header-2-1"),
        HeaderNode("Header 2.1.1", "3", "header-2-1-1"),
    ]

    assert create_toc(headers_list) == TocNode(
        entries=[
            TocEntryNode(
                value="Header 1",
                anchor="header-1",
                children=[
                    TocEntryNode(
                        value="Header 1.1",
                        anchor="header-1-1",
                        children=[],
                    ),
                    TocEntryNode(
                        value="Header 1.2",
                        anchor="header-1-2",
                        children=[],
                    ),
                ],
            ),
            TocEntryNode(
                value="Header 2",
                anchor="header-2",
                children=[
                    TocEntryNode(
                        value="Header 2.1",
                        anchor="header-2-1",
                        children=[
                            TocEntryNode(
                                value="Header 2.1.1",
                                anchor="header-2-1-1",
                                children=[],
                            ),
                        ],
                    ),
                ],
            ),
        ]
    )


def test_create_toc_orphan_nodes():
    headers_list = [
        HeaderNode("Header 1", "1", "header-1"),
        HeaderNode("Header 1.1.1", "3", "header-1-1-1"),
        HeaderNode("Header 1.2", "2", "header-1-2"),
    ]

    assert create_toc(headers_list) == TocNode(
        entries=[
            TocEntryNode(
                value="Header 1",
                anchor="header-1",
                children=[
                    TocEntryNode(
                        value="Header 1.1.1",
                        anchor="header-1-1-1",
                        children=[],
                    ),
                    TocEntryNode(
                        value="Header 1.2",
                        anchor="header-1-2",
                        children=[],
                    ),
                ],
            ),
        ]
    )



from mau.helpers import rematch
from mau.lexers.base_lexer import BaseLexer
from mau.lexers.base_lexer import TokenTypes as BLTokenTypes
from mau.text_buffer.context import Context


class TokenTypes:
    ARGUMENTS = "ARGUMENTS"
    BLOCK = "BLOCK"
    COMMAND = "COMMAND"
    COMMENT = "COMMENT"
    HEADER = "HEADER"
    HORIZONTAL_RULE = "HORIZONTAL_RULE"
    INCLUDE = "INCLUDE"
    LIST = "LIST"
    MULTILINE_COMMENT = "MULTILINE_COMMENT"
    TITLE = "TITLE"
    VARIABLE = "VARIABLE"


class MainLexer(BaseLexer):
    def _process_functions(self):
        return [
            self._process_multiline_comment,
            self._process_comment,
            self._process_horizontal_rule,
            self._process_block,
            self._process_command_or_directive,
            self._process_include,
            self._process_variable,
            self._process_arguments,
            self._process_title,
            self._process_list,
            self._process_header,
            # This is provided by BaseLexer
            self._process_text_line,
        ]

    def _process_multiline_comment(self):
        if self._current_line != "////":
            return None

        return self._create_tokens_from_line(TokenTypes.MULTILINE_COMMENT)

    def _process_comment(self):
        match = rematch(r"^(//.*)", self._current_line)

        if not match:
            return None

        tokens = [
            self._create_token_and_skip(TokenTypes.COMMENT, match.group(1)),
            self._create_token(BLTokenTypes.EOL),
        ]

        self._nextline()

        return tokens

    def _process_horizontal_rule(self):
        match = rematch(r"---$", self._current_line)

        if match is None:
            return None

        return [self._create_token_and_skip(TokenTypes.HORIZONTAL_RULE, "---")]

    def _process_block(self):
        if len(self._current_line) != 4:
            return None

        # Try to match the block delimiter
        match = rematch(r"^(.)\1{3}$", self._current_line)

        if match is None:
            return None

        return self._create_tokens_from_line(TokenTypes.BLOCK)

    def _run_directive(self, name, value):
        if name == "include":
            with open(value, encoding="utf-8") as included_file:
                text = included_file.read()
                text_buffer = self._text_buffer.__class__(text, Context(source=value))

                lexer = MainLexer(text_buffer)
                lexer.process()

                self.tokens.extend(lexer.tokens)

    def _process_command_or_directive(self):
        match = rematch(r"::([a-z0-9_#]+):(.*)?", self._current_line)

        if not match:
            return None

        command_name = match.group(1)

        if command_name.startswith("#"):
            # This is a lexer directive
            self._run_directive(command_name[1:], match.group(2))

            self._nextline()

            return []

        tokens = [
            self._create_token_and_skip(TokenTypes.COMMAND, "::"),
            self._create_token_and_skip(BLTokenTypes.TEXT, command_name),
            self._create_token_and_skip(BLTokenTypes.LITERAL, ":"),
        ]

        if match.group(2):
            tokens.append(
                self._create_token_and_skip(BLTokenTypes.TEXT, match.group(2))
            )

        tokens.append(self._create_token(BLTokenTypes.EOL))

        self._nextline()

        return tokens

    def _process_include(self):
        match = rematch(r"^<<( *)([a-z0-9_#\\]+):(.*)?", self._current_line)

        if not match:
            return None

        if not match.group(2):  # pragma: no cover
            return None

        tokens = [self._create_token_and_skip(TokenTypes.INCLUDE, "<<")]

        if match.group(1):
            tokens.append(
                self._create_token_and_skip(BLTokenTypes.WHITESPACE, match.group(1))
            )

        tokens.extend(
            [
                self._create_token_and_skip(BLTokenTypes.TEXT, match.group(2)),
                self._create_token_and_skip(BLTokenTypes.LITERAL, ":"),
            ]
        )

        if match.group(3):
            tokens.append(
                self._create_token_and_skip(BLTokenTypes.TEXT, match.group(3))
            )

        tokens.append(self._create_token(BLTokenTypes.EOL))

        self._nextline()

        return tokens

    def _process_variable(self):
        if not self._current_line.startswith(":"):
            return None

        match = rematch(r":([!a-zA-Z0-9\-_\.]+):(.*)?", self._current_line)

        if not match:  # pragma: no cover
            return None

        tokens = [
            self._create_token_and_skip(TokenTypes.VARIABLE, ":"),
            self._create_token_and_skip(BLTokenTypes.TEXT, match.group(1)),
            self._create_token_and_skip(BLTokenTypes.LITERAL, ":"),
        ]

        if match.group(2):
            tokens.append(
                self._create_token_and_skip(BLTokenTypes.TEXT, (match.group(2)))
            )

        tokens.append(self._create_token(BLTokenTypes.EOL))

        self._nextline()

        return tokens

    def _process_arguments(self):
        if not self._current_line.startswith("["):
            return None

        if not self._current_line.endswith("]"):
            return None

        line = self._current_line[1:-1]

        tokens = [
            self._create_token(TokenTypes.ARGUMENTS, "["),
            self._create_token(BLTokenTypes.TEXT, line),
            self._create_token(BLTokenTypes.LITERAL, "]"),
            self._create_token(BLTokenTypes.EOL),
        ]

        self._nextline()

        return tokens

    def _process_title(self):
        match = rematch(r"^\.( *)(.*)", self._current_line)

        if not match:
            return None

        if not match.group(2):  # pragma: no cover
            return None

        tokens = [
            self._create_token_and_skip(TokenTypes.TITLE, "."),
        ]

        if match.group(1):
            tokens.append(
                self._create_token_and_skip(BLTokenTypes.WHITESPACE, match.group(1))
            )

        tokens.extend(
            [
                self._create_token_and_skip(BLTokenTypes.TEXT, match.group(2)),
                self._create_token(BLTokenTypes.EOL),
            ]
        )

        self._nextline()

        return tokens

    def _process_list(self):
        match = rematch(r"^( *)([\*#]+)( +)(.*)", self._current_line)

        if not match:
            return None

        if not match.group(4):  # pragma: no cover
            return None

        tokens = []

        if match.group(1):
            tokens.append(
                self._create_token_and_skip(BLTokenTypes.WHITESPACE, match.group(1))
            )

        tokens.extend(
            [
                self._create_token_and_skip(TokenTypes.LIST, match.group(2)),
                self._create_token_and_skip(BLTokenTypes.WHITESPACE, match.group(3)),
                self._create_token_and_skip(BLTokenTypes.TEXT, match.group(4)),
                self._create_token(BLTokenTypes.EOL),
            ]
        )

        self._nextline()

        return tokens

    def _process_header(self):
        match = rematch("^(=+)( *)(.*)", self._current_line)

        if not match:
            return None

        if not match.group(3):
            return None

        tokens = [self._create_token_and_skip(TokenTypes.HEADER, match.group(1))]

        if match.group(2):
            tokens.append(
                self._create_token_and_skip(BLTokenTypes.WHITESPACE, match.group(2))
            )

        tokens.extend(
            [
                self._create_token_and_skip(BLTokenTypes.TEXT, match.group(3)),
                self._create_token(BLTokenTypes.EOL),
            ]
        )

        self._nextline()

        return tokens



import pytest
from mau.parsers.environment import Environment


def test_init():
    environment = Environment()

    assert environment.asdict() == {}


def test_init_with_content():
    environment = Environment({"var1": "value1", "var2": "value2"})

    assert environment.asdict() == {"var1": "value1", "var2": "value2"}


def test_init_with_content_and_environment():
    environment = Environment({"var1": "value1", "var2": "value2"}, "test")

    assert environment.asdict() == {"test": {"var1": "value1", "var2": "value2"}}


def test_flat_environment_default():
    environment = Environment()

    environment.setvar("var1", "value1")

    assert environment.getvar("var2", "defaultvalue") == "defaultvalue"


def test_set_and_get_variables_flat():
    environment = Environment()

    environment.setvar("var1", "value1")
    environment.setvar("var2", "value2")

    assert environment.getvar("var1") == "value1"
    assert environment.getvar("var2") == "value2"
    assert environment.asdict() == {"var1": "value1", "var2": "value2"}


def test_set_and_get_variables_namespace():
    environment = Environment()

    environment.setvar("namespace.var1", "value1")
    environment.setvar("namespace.var2", "value2")

    assert environment.getvar("namespace.var1") == "value1"
    assert environment.getvar("namespace.var2") == "value2"
    assert environment.asdict() == {"namespace": {"var1": "value1", "var2": "value2"}}


def test_set_and_get_variables_nested_namespace():
    environment = Environment()

    environment.setvar("name.space.var1", "value1")
    environment.setvar("name.space.var2", "value2")

    assert environment.getvar("name.space.var1") == "value1"
    assert environment.getvar("name.space.var2") == "value2"
    assert environment.asdict() == {
        "name": {"space": {"var1": "value1", "var2": "value2"}}
    }


def test_set_and_get_variables_no_default():
    environment = Environment()

    with pytest.raises(KeyError):
        environment.getvar("name.space.var2")


def test_update():
    environment = Environment()

    environment.update({"name1": "value1", "name2": "value2"})

    assert environment.asdict() == {"name1": "value1", "name2": "value2"}


def test_update_with_namespace():
    environment = Environment()

    environment.update({"name1": "value1", "name2": "value2"}, "test")

    assert environment.asdict() == {"test": {"name1": "value1", "name2": "value2"}}



from mau.helpers import rematch
from mau.lexers.base_lexer import BaseLexer, TokenTypes


class ArgumentsLexer(BaseLexer):
    def _process_functions(self):
        return [
            self._process_whitespace,
            self._process_literal,
            self._process_text,
        ]

    def _process_whitespace(self):
        match = rematch(r" +", self._tail)

        if not match:
            return None

        return [self._create_token_and_skip(TokenTypes.WHITESPACE, match.group())]

    def _process_literal(self):
        if self._current_char not in r'\=,"':
            return None

        return [self._create_token_and_skip(TokenTypes.LITERAL, self._current_char)]

    def _process_text(self):
        match = rematch(r'[^\\=," ]+', self._tail)

        if not match:  # pragma: no cover
            return None

        return [self._create_token_and_skip(TokenTypes.TEXT, match.group())]
